# coding: utf-8

import os

from io import StringIO
from unittest import TestCase

from citeproc.source.bibtex.bibparse import BibTeXParser, BibTeXDecodeError


TEST_BIB = os.path.join(os.path.dirname(__file__), 'test.bib')


class TestBibTeXParser(TestCase):
    def test_parse(self):
        file = StringIO(sample)
        bib = BibTeXParser(file)
        self.print_entries(bib)
        # TODO: perform useful checks

    def test_parse_file(self):
        test_bib = BibTeXParser(TEST_BIB, encoding='utf-8')
        self.print_entries(test_bib)

    def test_parse_file_bad_encoding(self):
        with self.assertRaises(BibTeXDecodeError):
            BibTeXParser(TEST_BIB)

    @staticmethod
    def print_entries(bib):
        for key, entry in bib.items():
            print(key)
            for name, value in entry.items():
                print('   {}: {}'.format(name, value))


# based on the sample BibTeX database by Xavier Décoret
# http://artis.imag.fr/~Xavier.Decoret/resources/xdkbibtex/bibtex_summary.html

sample = r"""
@Article(py03,
     author = {Xavier D\'ecoret},
     title  = "PyBiTex",
     year   = 2003
)

@Article{key03,
  title = "A {bunch {of} braces {in}} title"
}

@Article{key01a,
  author = "Simon {"}the {saint"} Templar",
}

@Article{key01b,
  title = "The history of @ sign"
}

Some {{comments} with unbalanced braces
....and a "commented" entry...

Book{landru21,
  author =	 {Landru, Henri D\'esir\'e},
  title =	 {A hundred recipes for you wife},
  publisher =	 {Culinary Expert Series},
  year =	 1921
}

..some other comments..before a valid entry...

@Book{steward03a,
  author =	 { Martha Steward },
  title =	 {Cooking behind bars},
  publisher =	 {Culinary Expert Series},
  year =	 2003
}

...and finally an entry commented by the use of the special @Comment entry type.

@Comment{steward03b,
  author =	 {Martha Steward},
  title =	 {Cooking behind bars},
  publisher =	 {Culinary Expert Series},
  year =	 2003
}

@Comment{
  @Book{steward03c,
    author =	 {Martha Steward},
    title =	 {Cooking behind bars},
    publisher =	 {Culinary Expert Series},
    year =	 2003
  }
}

@String(mar = "march")

@Book{sweig42,
  Author =	 { Stefan Sweig },
  title =	 { The impossible book },
  publisher =	 { Dead Poet Society},
  year =	 1942,
  month =        mar
}


@String {firstname = "Xavier"}
@String {lastname  = "Decoret"}
@String {email      = firstname # "." # lastname # "@imag.fr"}

@preamble{ "\newcommand{\noopsort}[1]{} "
        # "\newcommand{\printfirst}[2]{#1} "
        # "\newcommand{\singleletter}[1]{#1} "
        # "\newcommand{\switchargs}[2]{#2#1} " }

@INBOOK{inbook-minimal,
   author = "Donald E. Knuth",
   title = "Fundamental Algorithms",
   publisher = "Addison-Wesley",
   year = "{\noopsort{1973b}}1973",
   chapter = "1.2",
}
"""




import sys
import unicodedata

from . import BibliographySource, Reference
from . import Name, Date, DateRange, LiteralDate
from ..string import String, MixedString, NoCase
from .. import NAMES, DATES


class CiteProcJSON(BibliographySource):
    def __init__(self, json_data):
        for ref in json_data:
            ref_data = {}
            for key, value in ref.items():
                python_key = key.replace('-', '_')
                if python_key == 'id':
                    ref_key = str(value).lower()
                    continue
                elif python_key == 'type':
                    ref_type = value
                    continue
                elif python_key == 'key':
                    # conflicts with the ref_key, so ignore
                    continue
                if python_key == 'shortTitle':
                    python_key = 'title_short'

                if python_key in NAMES:
                    value = self.parse_names(value)
                elif python_key in DATES:
                    value = self.parse_date(value)
                elif key not in ('language', ):
                    value = self.parse_string(value)

                ref_data[python_key] = value
            self.add(Reference(ref_key, ref_type, **ref_data))

    start_tag = '<span class="nocase">'
    end_tag = '</span>'

    def parse_string(self, string):
        string = str(string)
        lower_string = string.lower()
        end = 0
        output = MixedString()
        try:
            while True:
                start = lower_string.index(self.start_tag, end)
                regular = string[end:start]
                if regular:
                    output += String(regular)
                end = lower_string.index(self.end_tag, start + 1)
                start += len(self.start_tag)
                no_case = string[start:end]
                output += NoCase(no_case)
                end += len(self.end_tag)
        except ValueError:
            regular = string[end:]
            if regular:
                output += String(regular)
        return output

    def parse_names(self, json_data):
        names = []
        for name_data in json_data:
            name = Name(**name_data)
            names.append(name)
        return names

    def parse_date(self, json_data):
        def parse_single_date(json_date):
            date_data = {}
            try:
                for i, part in enumerate(('year', 'month', 'day')):
                    date_data[part] = json_date[i]
            except IndexError:
                pass
            return date_data

        dates = []
        for json_date in json_data.get('date-parts', []):
            date = parse_single_date(json_date)
            if 'season' in json_data:
                date['season'] = json_data['season']
            dates.append(date)

        circa = json_data.get('circa', 0) != 0

        if len(dates) == 1:
            return Date(circa=circa, **dates[0])
        elif len(dates) > 1:
            return DateRange(begin=Date(**dates[0]), end=Date(**dates[1]),
                             circa=circa)
        elif 'literal' in json_data:
            return LiteralDate(json_data['literal'], circa=circa)
        else:
            return None




ARTICLE = 'article'
ARTICLE_JOURNAL = 'article-journal'
ARTICLE_MAGAZINE = 'article-magazine'
ARTICLE_NEWSPAPER = 'article-newspaper'
BILL = 'bill'
BOOK = 'book'
BROADCAST = 'broadcast'
CHAPTER = 'chapter'
ENTRY = 'entry'
ENTRY_DICTIONARY = 'entry-dictionary'
ENTRY_ENCYCLOPEDIA = 'entry-encyclopedia'
FIGURE = 'figure'
GRAPHIC = 'graphic'
INTERVIEW = 'interview'
LEGAL_CASE = 'legal_case'
LEGISLATION = 'legislation'
MANUSCRIPT = 'manuscript'
MAP = 'map'
MOTION_PICTURE = 'motion_picture'
MUSICAL_SCORE = 'musical_score'
PAMPHLET = 'pamphlet'
PAPER_CONFERENCE = 'paper-conference'
PATENT = 'patent'
PERSONAL_COMMUNICATION = 'personal_communication'
POST = 'post'
POST_WEBLOG = 'post-weblog'
REPORT = 'report'
REVIEW = 'review'
REVIEW_BOOK = 'review-book'
SONG = 'song'
SPEECH = 'speech'
THESIS = 'thesis'
TREATY = 'treaty'
WEBPAGE = 'webpage'



from citeproc.string import String, NoCase


a = String('a ')
b = String(' b ')
nocase = NoCase('NoCase')
nocase2 = NoCase('NoCaseTwo')

c = a + nocase + b + nocase2

print(c)
print(c.upper())
print(c.lower())

print('test' + b)
print('test ' + nocase)
print(a + 'test')
print(nocase + ' test')

print('test ' + c)
print(c + ' test')




try:
    from html import escape
except ImportError:
    from cgi import escape


def preformat(text):
    return escape(str(text), quote=False)


class TagWrapper(str):
    tag = None
    attributes = None

    @classmethod
    def _wrap(cls, text):
        if cls.attributes:
            attrib = ' ' + ' '.join(['{}="{}"'.format(key, value)
                                     for key, value in cls.attributes.items()])
        else:
            attrib = ''
        return '<{tag}{attrib}>{text}</{tag}>'.format(tag=cls.tag,
                                                      attrib=attrib,text=text)

    def __new__(cls, text):
        return super(TagWrapper, cls).__new__(cls, cls._wrap(text))


class Italic(TagWrapper):
    tag = 'i'


class Oblique(Italic):
    pass


class Bold(TagWrapper):
    tag = 'b'


class Light(TagWrapper):
    tag = 'l'


class Underline(TagWrapper):
    tag = 'u'


class Superscript(TagWrapper):
    tag = 'sup'


class Subscript(TagWrapper):
    tag = 'sub'


class SmallCaps(TagWrapper):
    tag = 'span'
    attributes = {'style': 'font-variant:small-caps;'}



# coding: utf-8

from unittest import TestCase

from citeproc.source.bibtex.latex import parse_latex, substitute_ligatures


class TestLatex(TestCase):
    # from "A TEX Primer for Scientists" (sections 5.2 and 5.3)
    # by Stanley A. Sawyer, Steven G. Krantz
    MACROS = [("Æsop's fables", r"\AE sop's fables"),
              ('Æ is a Latin ligature', r"\AE\ is a Latin ligature"),
              ('A TeX expert is a TeXnician.',
                    r"A \TeX\ expert is a \TeX nician."),
              ('This book is ©ed.', r"This book is \copyright ed."),
              ('Umeå, Sweden', r"Ume\aa, Sweden"),
              ('Your serial number is Å1102.',
                    r"Your serial number is \AA1102."),
              ('Your serial number is Å1102.',
                    r"Your serial number is \AA  1102."),
              ('$22.50 plus $ 4.07 tax.', r"\$22.50 plus \$ 4.07 tax."),
              ('Crève Cœur, Missouri', r"Cr\`eve C\oe ur, Missouri"),
              ('Pëtr, meet Françoise', r"P\"etr, meet Fran\c coise"),

              ('Pëtr Pëtr Pëtr', r'P\" etr P\"  etr ' 'P\\"\tetr'),
              ]

    ACCENT_MACROS = [('très élevé', r"tr\`es \'elev\'e"),
                     ('übel Föhn', r'\"ubel F\"ohn'),
                     ('Andō Tokutarō', r"And\=o Tokutar\=o"),
                     ('García Lorca', r"Garc\'\i a Lorca"),
                     ('naïve', r'na\"\i ve'),

                     ('hát hàt hȧt hät hāt', r"h\'at h\`at h\.at h\"at h\=at"),
                     ('hãt  hât  ha̧t  ha̱t', r"h\~at  h\^at  h\c at  h\b at"),
                     ('hạt  hǎt  ha̋t', "h\\d at  h\\v at  h\\H at"),
                     ('hăt  ha͡at', "h\\u at  h\\t aat"),

                     ('øre', r"\o re"),
                     ('Øre', r"\O re"),
                     ('łódka', r"\l\'odka"),
                     ('Łodz', r"\L odz"),
                     ('Altstraße', r"Altstra\ss e"),
                     ('æsthete', r"\ae sthete"),
                     ('Ålm', r"\AA lm"),
                     # ('', r"{\it\$}5"),
                     ('Œuvre', r"\OE uvre"),
                     ]

    # assorted string with macros
    ASSORTED = [('Escobar, María José', r"Escobar, Mar{\'\i}a Jos{\'e}"),
                ('Escobar, María-José', r"Escobar, Mar\'{\i}a-Jos\'{e}")]

    WARNINGS = [('Åke José Édouard Gödel',
                 r"\AA{ke} {Jos{\'{e}} {\'{E}douard} G{\"o}del",
                 "Unbalanced parenthesis")]

    MATH = [(r'An $O(n \log n / \! \log\log n)$ Sorting Algorithm',
                  r"An $O(n \log n / \! \log\log n)$ Sorting Algorithm")]

    def test_parse_latex(self):
        for reference, string in self.MACROS:
            self.assertEqual(reference, parse_latex(string))
        for reference, string in self.ACCENT_MACROS:
            self.assertEqual(reference, parse_latex(string))
        for reference, string in self.ASSORTED:
            self.assertEqual(reference, parse_latex(string))
        for reference, string, warning in self.WARNINGS:
            with self.assertWarns(UserWarning) as record:
                self.assertEqual(reference, parse_latex(string))
            message, = (rec.message.args[0] for rec in record.warnings)
            assert warning in message
        for reference, string in self.MATH:
            self.assertEqual(reference, parse_latex(string))

    LIGATURES = [('¿Que pasa?', "?`Que pasa?"),
                 ('¡Que!', "!`Que!")]

    def test_substitute_ligatures(self):
        for reference, string in self.LIGATURES:
            self.assertEqual(reference, substitute_ligatures(string))




import re
import unicodedata

from warnings import warn

from ...types import (ARTICLE, ARTICLE_JOURNAL, BOOK, CHAPTER, MANUSCRIPT,
                      PAMPHLET, PAPER_CONFERENCE, REPORT, THESIS)
from ...string import String, MixedString, NoCase
from .. import BibliographySource, Reference, Name, Date, DateRange
from .bibparse import BibTeXParser
from .latex import parse_latex
from .latex.macro import NewCommand, Macro


class BibTeX(BibliographySource):
    fields = {'abstract': 'abstract',
              'address': 'publisher_place',
              'annote': 'annote',
              'author': 'author',
              'booktitle': 'container_title',
              'chapter': 'chapter_number',
              'doi': 'DOI',
              'edition': 'edition',
              'editor': 'editor',
#              'howpublished': None,
#              'institution': None,
              'journal': 'container_title',
#              'month': None,
              'note': 'note',
              'number': 'issue',
#              'organization': None,
              'pages': 'page',
              'pmid': 'PMID',
              'publisher': 'publisher',
#              'school': None,
              'series': 'collection_title',
              'title': 'title',
#              'type': None,
#              'year': None,
              'volume': 'volume',

              # non-standard fields
              'isbn': 'ISBN',
              'issn': 'ISSN'}

    types = {# standard entry types
             'article': ARTICLE_JOURNAL,
             'book': BOOK,
             'booklet': PAMPHLET,
             'conference': PAPER_CONFERENCE,
             'inbook': CHAPTER,
             'incollection': ARTICLE_JOURNAL,
             'inproceedings': PAPER_CONFERENCE,
             'manual': BOOK,
             'mastersthesis': THESIS,
             'misc': ARTICLE,
             'phdthesis': THESIS,
             'proceedings': BOOK,
             'techreport': REPORT,
             'unpublished': MANUSCRIPT,

             # non-standard entry types
             'thesis': THESIS,
             'report': REPORT,
             }

    def __init__(self, filename, encoding='ascii'):
        bibtex_database = BibTeXParser(filename, encoding)
        self.preamble_macros = {}
        parse_latex(bibtex_database.preamble,
                    {'newcommand': NewCommand(self.preamble_macros),
                     'mbox': Macro(1, '{0}'),
                     'cite': Macro(1, 'CITE({0})')})
        for key, entry in bibtex_database.items():
            self.add(self.create_reference(key, entry))

    def _bibtex_to_csl(self, bibtex_entry):
        csl_dict = {}
        for field, value in bibtex_entry.items():
            try:
                value = value.strip()
            except AttributeError:
                pass
            try:
                csl_field = self.fields[field]
            except KeyError:
                if field not in ('year', 'month', 'filename'):
                    warn("Unsupported BibTeX field '{}'".format(field))
                continue
            if field in ('number', 'volume'):
                try:
                    value = int(value)
                except ValueError:
                    pass
            elif field == 'pages':
                value = self._bibtex_to_csl_pages(value)
            elif field in ('author', 'editor'):
                value = [name for name in self._parse_author(value)]
            else:
                try:
                    value = self._parse_string(value)
                except TypeError:
                    value = str(value)
            csl_dict[csl_field] = value
        return csl_dict

    @staticmethod
    def _bibtex_to_csl_pages(value):
        value = value.replace(' ', '')
        if '-' in value:
            try:
                first, last = value.split('--')
            except ValueError:
                first, last = value.split('-')
            pages = '-'.join((first, last))
        else:
            pages = value[:-1] if value.endswith('+') else value
        return pages

    def _bibtex_to_csl_date(self, bibtex_entry):
        if 'month' in bibtex_entry:
            begin_dict, end_dict = self._parse_month(bibtex_entry['month'])
        else:
            begin_dict, end_dict = {}, {}
        if 'year' in bibtex_entry:
            begin_dict['year'], end_dict['year'] \
                = self._parse_year(bibtex_entry['year'])
        if not begin_dict:
            return None
        if begin_dict == end_dict:
            return Date(**begin_dict)
        else:
            return DateRange(begin=Date(**begin_dict), end=Date(**end_dict))

    def _parse_year(self, year):
        try:
            year_str = parse_latex(year, self.preamble_macros)
        except TypeError:
            year_str = str(year)
        if EN_DASH in year_str:
            begin_year, end_year = year_str.split(EN_DASH)
            begin_len, end_len = len(begin_year), len(end_year)
            if end_len < begin_len:
                end_year = begin_year[:begin_len - end_len] + end_year
        else:
            begin_year = end_year = int(year_str)
        return begin_year, end_year

    MONTHS = ('jan', 'feb', 'mar', 'apr', 'may', 'jun',
              'jul', 'aug', 'sep', 'oct', 'nov', 'dec')
    RE_DAY = r'(?P<day>\d+)'
    RE_MONTH = r'(?P<month>\w+)'

    @staticmethod
    def _parse_month(month):
        def month_name_to_index(name):
            try:
                return BibTeX.MONTHS.index(name[:3].lower()) + 1
            except ValueError:
                return int(name)

        begin = {}
        end = {}
        month = month.strip()
        month = month.replace(', ', '-')
        if month.isdecimal():
            begin['month'] = end['month'] = month
        elif month.replace('-', '').isalpha():
            if '-' in month:
                begin['month'], end['month'] = month.split('-')
            else:
                begin['month'] = end['month'] = month
        else:
            m = re.match(BibTeX.RE_DAY + '[ ~]*' + BibTeX.RE_MONTH, month)
            if m is None:
                m = re.match(BibTeX.RE_MONTH + '[ ~]*' + BibTeX.RE_DAY, month)
            begin['day'] = end['day'] = int(m.group('day'))
            begin['month'] = end['month'] = m.group('month')
        begin['month'] = month_name_to_index(begin['month'])
        end['month'] = month_name_to_index(end['month'])
        return begin, end

    def _parse_string(self, title):
        def make_string(string, top_level_group=False):
            unlatexed = parse_latex(string, self.preamble_macros)
            fixed_case = top_level_group and not string.startswith('\\')
            string_cls = NoCase if fixed_case else String
            return string_cls(unlatexed)

        output = MixedString()
        level = 0
        string = ''
        for char in title:
            if char == '{':
                if level == 0:
                    if string:
                        output += make_string(string)
                        string = ''
                level += 1
            elif char == '}':
                level -= 1
                if level == 0:
                    output += make_string(string, True)
                    string = ''
            else:
                string += char
        if level != 0:
            raise SyntaxError('Non-matching braces in "{}"'.format(title))
        if string:
            output += make_string(string)
        return output

    def _parse_author(self, authors):
        csl_authors = []
        for author in split_names(authors):
            first, von, last, jr = parse_name(author)
            csl_parts = {}
            for part, csl_label in [(first, 'given'),
                                    (von, 'non-dropping-particle'),
                                    (last, 'family'),
                                    (jr, 'suffix')]:
                if part is not None:
                    csl_parts[csl_label] = parse_latex(part,
                                                       self.preamble_macros)
            name = Name(**csl_parts)
            csl_authors.append(name)
        return csl_authors

    def create_reference(self, key, bibtex_entry):
        csl_type = self.types[bibtex_entry.document_type]
        csl_fields = self._bibtex_to_csl(bibtex_entry)
        csl_date = self._bibtex_to_csl_date(bibtex_entry)
        if csl_date:
            csl_fields['issued'] = csl_date
        return Reference(key, csl_type, **csl_fields)


# BibTeX name handling
#
# references
#  - BibTeXing by Oren Patashnik (Feb 8, 1988), 4. Helpful Hints, item 18
#    (BibTeX 0.99d - http://www.ctan.org/tex-archive/biblio/bibtex/base/btxdoc.pdf)
#  - A summary of BibTex by Xavier Décoret
#    (http://maverick.inria.fr/~Xavier.Decoret/resources/xdkbibtex/bibtex_summary.html)
#  - Tame the BeaST by Nicolas Markey
#    (http://tug.ctan.org/info/bibtex/tamethebeast/ttb_en.pdf)

AND = ' and '

def split_names(string):
    """Split a string of names separated by 'and' into a list of names."""
    brace_level = 0
    names = []
    last_index = 0
    for i in range(len(string)):
        char = string[i]
        if brace_level == 0 and string[i:].startswith(AND):
            names.append(string[last_index:i])
            last_index = i + len(AND)
        elif char == '{':
            brace_level += 1
        elif char == '}':
            brace_level -= 1
    last_name = string[last_index:]
    if last_name:
        names.append(last_name)
    return names


def parse_name(name):
    """Parse a BibTeX name string and split it into First, von, Last and Jr
    parts.
    """
    parts = split_name(name)
    if len(parts) == 1:       # First von Last
        first_von_last, = parts
        index = 0
        first, jr = [], []
        for word in first_von_last[:-1]:
            if is_capitalized(word) not in (True, None):
                break
            first.append(word)
            index += 1
        von_last = first_von_last[index:]
    elif len(parts) == 2:     # von Last, First
        jr = []
        von_last, first = parts
    elif len(parts) == 3:     # von Last, Jr, First
        von_last, jr, first = parts
    von, last = split_von_last(von_last)
    join = ' '.join
    return join(first) or None, join(von) or None, join(last), join(jr) or None


def split_name(name):
    """Split a name in into parts delimited by commas (at brace-level 0), and
    each part into words.

    Returns a list of of lists of words.
    """
    brace_level = 0
    parts = []
    current_part = []
    word = ''
    for char in name:
        if char in ' \t,':
            if brace_level == 0:
                if word:
                    current_part.append(word)
                    word = ''
                if char == ',':
                    parts.append(current_part)
                    current_part = []
                continue
        elif char == '{':
            brace_level += 1
        elif char == '}':
            brace_level -= 1
        word += char
    if word:
        current_part.append(word)
        parts.append(current_part)
    return parts


def is_capitalized(string):
    """Check if a BibTeX substring is capitalized.

    A string can be "case-less", in which case `None` is returned.
    """
    brace_level = 0
    special_char = False
    for char, next_char in lookahead_iter(string):
        if (brace_level == 0 or special_char) and char.isalpha():
            return char.isupper()
        elif char == '{':
            brace_level += 1
            if brace_level == 1 and next_char == '\\':
                special_char = True
        elif char == '}':
            brace_level -= 1
            if brace_level == 0:
                special_char = False
    return None  # case-less


def split_von_last(words):
    """Split "von Last" name into von and Last parts."""
    if len(words) > 1 and is_capitalized(words[0]) is False:
        for j, word in enumerate(reversed(words[:-1])):
            if is_capitalized(word) not in (True, None):
                return words[:-j - 1], words[-j - 1:]
    return [], words


def lookahead_iter(iterable):
    """Iterator that also yields the next item along with each item. The next
    item is `None` when yielding the last item.
    """
    items = iter(iterable)
    item = next(items)
    for next_item in items:
        yield item, next_item
        item = next_item
    yield item, None


EN_DASH = unicodedata.lookup('EN DASH')




def escape(text):
    text = text.replace('*', r'\*')
    text = text.replace('`', r'\`')
    return text


def preformat(text):
    return escape(str(text))


class RoleWrapper(str):
    role = None

    @classmethod
    def _wrap(cls, text):
        return ':{role}:`{text}`'.format(role=cls.role, text=text)

    def __new__(cls, text):
        return super(RoleWrapper, cls).__new__(cls, cls._wrap(text))


class Italic(RoleWrapper):
    role = 'emphasis'


class Oblique(Italic):
    pass


class Bold(RoleWrapper):
    role = 'strong'


Light = str
Underline = str


class Superscript(RoleWrapper):
    role = 'superscript'


class Subscript(RoleWrapper):
    role = 'subscript'


SmallCaps = str




# http://maverick.inria.fr/~Xavier.Decoret/resources/xdkbibtex/bibtex_summary.html
# http://www.lsv.ens-cachan.fr/~markey/bibla.php?lang=en

class BibTeXEntry(dict):
    def __init__(self, document_type, attributes):
        super(BibTeXEntry, self).__init__(attributes)
        self.document_type = document_type


class BibTeXDecodeError(Exception):
    """Exception raised when the encoding passed to BibTeXParser cannot decode
    the BibTeX database file."""
    def __init__(self, decode_error, line_number):
        msg = ("'{}' decode error on line {}: {}. Please specify the BibTeX's "
               "database character encoding when instantiating BibTeXParser."
               .format(decode_error.encoding, line_number, decode_error.reason))
        super(BibTeXDecodeError, self).__init__(msg)
        self.line_number = line_number
        self.decode_error = decode_error


class BibTeXParser(dict):
    standard_variables = {'jan': 'January',
                          'feb': 'February',
                          'mar': 'March',
                          'apr': 'April',
                          'may': 'May',
                          'jun': 'June',
                          'jul': 'July',
                          'aug': 'August',
                          'sep': 'September',
                          'oct': 'October',
                          'nov': 'November',
                          'dec': 'December'}

    def __init__(self, file_or_filename, encoding='ascii'):
        try:
            self.file = open(file_or_filename, 'rt', encoding=encoding)
        except TypeError:
            self.file = file_or_filename
        self.variables = {}
        self.preamble = ''
        self._parse(self.file)
        self.file.close()

    def _parse(self, file):
        while True:
            try:
                entry = self._parse_entry(file)
                if entry is not None:
                    entry_type, key, attributes = entry
                    self[key] = BibTeXEntry(entry_type, attributes)
            except UnicodeDecodeError as decode_error:
                # I'm not convinced the following is robust
                safe_part = decode_error.object[:decode_error.start]
                line_number = len(safe_part.splitlines())
                offset = file.tell() - len(decode_error.object)
                if offset > 0:
                    file.seek(0)
                    line_number += file.read(offset).count(file.newlines)
                raise BibTeXDecodeError(decode_error, line_number)
            except EOFError:
                break

    def _parse_entry(self, file):
        while True:
            char = file.read(1)
            if char == '@':
                break
            elif char == '':
                raise EOFError
        entry_type = ''
        while True:
            char = file.read(1)
            if char in '{(':
                sentinel = '}' if char == '{' else ')'
                break
            entry_type += char
            if entry_type.lower().startswith('comment'):
                self._jump_to_next_line(file)
                return None
        entry_type = entry_type.strip().lower()
        if entry_type == 'string':
            name = self._parse_name(file)
            value = self._parse_value(file)
            self.variables[name] = value
            assert self._eat_whitespace(file) == sentinel
            return None
        elif entry_type == 'preamble':
            self.preamble += self._parse_value(file)
            assert self._eat_whitespace(file) == sentinel
            return None
        key = self._parse_key(file)
        entry = {}
        while True:
            name = self._parse_name(file)
            value = self._parse_value(file)
            entry[name] = value
            char = self._eat_whitespace(file)
            if char != ',':
                if char != sentinel:
                    assert char in ' \t\n\r'
                    assert self._eat_whitespace(file) == sentinel
                break
            else:
                restore_point = file.tell()
                if self._eat_whitespace(file) == sentinel:
                    break
                file.seek(restore_point)
        return entry_type.strip().lower(), key, entry

    def _parse_key(self, file):
        key = ''
        char = file.read(1)
        while char != ',':
            key += char
            char = file.read(1)
            if not char:
                raise ValueError('End of file while parsing key')
        return key.strip().lower()

    def _parse_name(self, file):
        name = ''
        char = self._eat_whitespace(file)
        while True:
            if char == '=':
                break
            name += char
            char = file.read(1)
        return name.strip().lower()

    def _parse_value(self, file):
        char = self._eat_whitespace(file)
        if char in '{"':
            value = self._parse_string(file, char)
        elif char.isalpha():
            value = self._parse_variable(file, char)
        else:
            value = self._parse_integer(file, char)

        restore_position = file.tell()
        char = self._eat_whitespace(file)
        if char == '#':
            value += self._parse_value(file)
        else:
            file.seek(restore_position)
        return value

    def _parse_string(self, file, opening_character):
        closing_character = '"' if opening_character == '"' else '}'
        string = ''
        depth = 0
        while True:
            char = file.read(1)
            if not char:
                raise ValueError('End of file while parsing string value')
            if char == '{':
                depth += 1
            elif depth == 0 and char == closing_character:
                break
            elif char == '}':
                depth -= 1
            string += char
        return string

    def _parse_variable(self, file, char):
        key = ''
        restore_point = file.tell()
        while char.isalnum() or char in '-_':
            key += char
            restore_point = file.tell()
            char = file.read(1)
        file.seek(restore_point)
        if key.lower() in self.variables:
            value = self.variables[key.lower()]
        else:
            value = self.standard_variables[key.lower()]
        return value

    def _parse_integer(self, file, char):
        integer = ''
        restore_point = file.tell()
        while char.isdigit():
            integer += char
            restore_point = file.tell()
            char = file.read(1)
        file.seek(restore_point)
        return int(integer)

    # TODO: rename to next_token?
    def _eat_whitespace(self, file):
        char = file.read(1)
        while char in ' \t\n\r':
            char = file.read(1)
        return char

    def _jump_to_next_line(self, file):
        char = ''
        while char != '\n':
            restore_point = file.tell()
            char = file.read(1)
        file.seek(restore_point)

    def _split_name(self, name):
        pass




import os
import sys
import warnings
from optparse import OptionParser

# The references are parsed from a BibTeX database, so we import the
# corresponding parser.
from citeproc.source.bibtex import BibTeX

# Import the citeproc-py classes we'll use below.
from citeproc import frontend
from citeproc import CitationStylesStyle, CitationStylesBibliography
from citeproc import formatter
from citeproc import Citation, CitationItem


def main():
    usage = \
"""usage: %prog [options] <bib_file>'

Output a formatted bibliography from bibtex file <bib_file> to stdout

You might want to put the path to your CSL styles in the `CSL_STYLES_PATH`
environment variable with something like (bash shell)::

    export CSL_STYLES_PATH=$HOME/dev_trees/styles
"""
    parser = OptionParser(usage)
    parser.add_option('-s', '--style', dest='style', default='harvard1',
                      help='style file', metavar='STYLE_FILE')
    parser.add_option('-f', '--format', dest='format', default='rst',
                      help='output type', metavar='FORMAT')
    (options, args) = parser.parse_args()
    if len(args) != 1:
        parser.print_usage()
        sys.exit(1)

    output_format = options.format
    try:
        output_format = getattr(formatter, options.format)
    except AttributeError:
        available_formatters = (attr for attr in dir(formatter)
                                if not attr.startswith('__'))
        print('style should be one of: ' + ', '.join(available_formatters))
        sys.exit(1)

    # Get styles path out of the environment variable
    env_path = os.environ.get('CSL_STYLES_PATH')
    if env_path is not None:
        frontend.STYLES_PATH = env_path

    # Silence lots of warnings
    warnings.simplefilter('ignore')
    # Read bibliography
    bib_source = BibTeX(args[0])
    # load a CSL style (full path or name of style in style path)
    bib_style = CitationStylesStyle(options.style)
    # Create the citeproc-py bibliography, passing it the:
    # * CitationStylesStyle,
    # * BibliographySource (BibTeX in this case), and
    # * a formatter (plain, html, or rst)
    bibliography = CitationStylesBibliography(bib_style, bib_source,
                                              output_format)
    # Processing citations in a document need to be done in two passes as for
    # some CSL styles, a citation can depend on the order of citations in the
    # bibliography and thus on citations following the current one.
    # For this reason, we first need to register all citations with the
    # CitationStylesBibliography.
    # Just read every one in key order
    for name in sorted(bib_source):
        bibliography.register(Citation([CitationItem(name)]))
    # In the second pass, CitationStylesBibliography can generate citations.
    # CitationStylesBibliography.cite() requires a callback function to be
    # passed along to be called in case a CitationItem's key is not present in
    # the bibliography.
    for item in bibliography.bibliography():
        print(str(item))


if __name__ == '__main__':
    main()




import os

from warnings import warn

from lxml import etree

from . import SCHEMA_PATH, LOCALES_PATH, STYLES_PATH
from .model import CitationStylesElement
from .formatter import html


class CitationStylesXML(object):
    def __init__(self, f, validate=True):
        lookup = etree.ElementNamespaceClassLookup()
        namespace = lookup.get_namespace('http://purl.org/net/xbiblio/csl')
        namespace[None] = CitationStylesElement
        namespace.update(dict([(cls.__name__.replace('_', '-').lower(), cls)
                               for cls in CitationStylesElement.__subclasses__()]))

        self.parser = etree.XMLParser(remove_comments=True, encoding='UTF-8',
                                      no_network=True)
        self.parser.set_element_class_lookup(lookup)
        self.xml = etree.parse(f, self.parser)#, base_url=".")
        if validate:
            self.schema = etree.RelaxNG(etree.parse(SCHEMA_PATH))
            if not self.schema.validate(self.xml):
                err = self.schema.error_log
                #raise Exception("XML file didn't pass schema validation:\n%s" % err)
                warn("XML file didn't pass schema validation:\n%s" % err)
                # TODO: proper error reporting
        self.root = self.xml.getroot()


class CitationStylesLocale(CitationStylesXML):
    def __init__(self, locale, validate=True):
        locale_path = os.path.join(LOCALES_PATH, 'locales-{}.xml'.format(locale))
        try:
            super(CitationStylesLocale, self).__init__(locale_path,
                                                       validate=validate)
        except IOError:
            raise ValueError("'{}' is not a known locale".format(locale))


class CitationStylesStyle(CitationStylesXML):
    def __init__(self, style, locale=None, validate=True):
        try:
            if not os.path.exists(style):
                style = os.path.join(STYLES_PATH, '{}.csl'.format(style))
        except TypeError:
            pass
        try:
            super(CitationStylesStyle, self).__init__(
                style, validate=validate)
        except IOError:
            raise ValueError("'{}' is not a known style".format(style))
        if locale is None:
            locale = self.root.get('default-locale', 'en-US')
        self.root.set_locale_list(locale, validate=validate)

    def has_bibliography(self):
        return self.root.bibliography is not None

    def render_citation(self, citation, cites, callback=None):
        return self.root.citation.render(citation, cites, callback)

    def sort_bibliography(self, citation_items):
        return self.root.bibliography.sort(citation_items)

    def render_bibliography(self, citation_items):
        return self.root.bibliography.render(citation_items)


class CitationStylesBibliography(object):
    def __init__(self, style, source, formatter=html):
        self.style = style
        self.source = source
        self.formatter = self.style.root.formatter = formatter
        self.keys = []
        self.items = []
        self._cites = []

    def register(self, citation, callback=None):
        citation.bibliography = self
        for item in citation.cites:
            if item.key in self.source:
                if item.key not in self.keys:
                    self.keys.append(item.key)
                    self.items.append(item)
            elif callback is not None:
                callback(item)

    def sort(self):
        self.items = self.style.sort_bibliography(self.items)
        self.keys = [item.key for item in self.items]

    def cite(self, citation, callback):
        return self.style.render_citation(citation, self._cites, callback)

    def bibliography(self):
        return self.style.render_bibliography(self.items)




from functools import wraps, reduce


def discard_empty_other(method):
    """Decorator for addition operator methods that returns the object itself if
    `other` is the empty string."""
    @wraps(method)
    def wrapper(obj, other):
        if other == '':
            return obj
        else:
            return method(obj, other)
    return wrapper


class String(str):
    @discard_empty_other
    def __radd__(self, other):
        return MixedString([other]).__add__(self)

    @discard_empty_other
    def __add__(self, other):
        return MixedString([self]).__add__(other)

    def __iadd__(self, other):
        return self.__add__(other)

    def replace(self, *args, **kwargs):
        return self.__class__(super(String, self).replace(*args, **kwargs))

    def rstrip(self, *args, **kwargs):
        return self.__class__(super(String, self).rstrip(*args, **kwargs))

    def lower(self):
        return self.__class__(super(String, self).lower())

    def upper(self):
        return self.__class__(super(String, self).upper())

    def soft_lower(self):
        return self.lower()

    def soft_upper(self):
        return self.upper()

    def capitalize_first(self):
        return self.__class__(self[0].upper() + self[1:])

    def words(self):
        for word in self.split():
            yield self.__class__(word)


class MixedString(list):
    @discard_empty_other
    def __add__(self, other):
        super_obj = super(MixedString, self)
        try:
            return self.__class__(super_obj.__add__(other))
        except TypeError:
            return self.__class__(super_obj.__add__(MixedString([other])))

    @discard_empty_other
    def __radd__(self, other):
        return self.__class__([other]).__add__(self)

    def __iadd__(self, other):
        return self.__add__(other)

    def __str__(self):
        return ''.join(map(str, self))

    def __getitem__(self, index):
        return str(self)[index]

    def replace(self, *args):
        return self.__class__([string.replace(*args) for string in self])

    def translate(self, table):
        return self.__class__([string.translate(table) for string in self])

    def lower(self):
        return self.__class__([string.lower() for string in self])

    def upper(self):
        return self.__class__([string.upper() for string in self])

    def title(self):
        return self.__class__([string.title() for string in self])

    def capitalize_first(self):
        self_iter = iter(self)
        output = [next(self_iter).capitalize_first()]
        output += [string for string in self_iter]
        return self.__class__(output)

    def isupper(self):
        return all(string.isupper() for string in self)

    def split(self, *args, **kwargs):
        return str(self).split(*args, **kwargs)

    def rstrip(self, *args, **kwargs):
        rev_iter = reversed(self)
        output = [next(rev_iter).rstrip(*args, **kwargs)]
        output += [string for string in rev_iter]
        return self.__class__(reversed(output))

    def words(self):
        for string in self:
            for word in string.words():
                yield word


class NoCase(String):
    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, str(self))

    def soft_lower(self):
        return self

    def soft_upper(self):
        return self

    def capitalize_first(self):
        return self


def join(items, delimiter=''):
    return reduce(lambda a, b: a + delimiter + b, items)




# http://sourceforge.net/mailarchive/message.php?msg_id=25355232

# http://dret.net/bibconvert/tex2unicode

from warnings import warn

from .. import VARIABLES


class CustomDict(dict):
    def __init__(self, args, required=set(), optional=set(), required_or=[]):
        passed_keywords = set(args.keys())
        missing = required - passed_keywords
        if missing:
            raise TypeError('The following required arguments are missing: ' +
                            ', '.join(missing))
        required_or_merged = set()
        for required_options in required_or:
            if not passed_keywords & required_options:
                raise TypeError('Require at least one of: ' +
                                ', '.join(required_options))
            required_or_merged |= required_options
        unsupported = passed_keywords - required - optional - required_or_merged
        if unsupported:
            cls_name = self.__class__.__name__
            warn('The following arguments for {} are '.format(cls_name) +
                 'unsupported: ' + ', '.join(unsupported))
        self.update(args)

    def __setattr__(self, name, value):
        self[name] = value

    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, key):
        try:
            return super(CustomDict, self).__getitem__(key)
        except KeyError:
            raise VariableError


class Reference(CustomDict):
    def __init__(self, key, type, **args):
        self.key = key
        self.type = type
        #required_or = [set(csl.VARIABLES)]
        optional = ({'uri', 'container_uri', 'contributor', 'date'} |
                    set(VARIABLES))
        super(Reference, self).__init__(args, optional=optional)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.key)


class VariableError(Exception):
    pass


class Name(CustomDict):
    def __init__(self, **args):
        if 'literal' in args:
            required, optional = {'literal'}, set()
        else:
            required = {'family'}
            optional = {'given', 'dropping-particle', 'non-dropping-particle',
                        'suffix'}
        super(Name, self).__init__(args, required, optional)

    def parts(self):
        if 'literal' in self:
            return (None, self['literal'], None, None, None)
        else:
            return (self.get('given'), self.get('family'),
                    self.get('dropping-particle'),
                    self.get('non-dropping-particle'), self.get('suffix'))


class DateBase(CustomDict):
    def __init__(self, args, required=set(), optional=set()):
        optional = {'circa'} | optional
        super(DateBase, self).__init__(args, required, optional)
        # defaults
        if 'circa' not in self:
            self['circa'] = False


class Date(DateBase):
    def __init__(self, **args):
        required = {'year'}
        optional = {'month', 'day', 'season'}
        if 'day' in args and 'month' not in args:
            raise TypeError('When specifying the day, you should also specify '
                            'the month')
        args = {key: int(value) for key, value in args.items()}
        super(Date, self).__init__(args, required, optional)

    def sort_key(self):
        year = self.year
        month = self.get('month', 0)
        day = self.get('day', 0)
        return '{:05}{:02}{:02}'.format(year + 10000, month, day)

    def is_nil(self):
        return (self.year == 0 and self.get('month', 0) == 0 and
                self.get('day', 0) == 0)


class LiteralDate(DateBase):
    def __init__(self, text, **args):
        self.text = text
        super(LiteralDate, self).__init__(args)

    def sort_key(self):
        return self.text


class DateRange(DateBase):
    def __init__(self, **args):
        required = {'begin'}
        optional = {'end'}
        super(DateRange, self).__init__(args, required, optional)

    def sort_key(self):
        begin = self.begin.sort_key()
        end = self.get('end', Date(year=0)).sort_key()
        return begin + '-' + end

    def __eq__(self, other):
        # TODO: for sorting
        raise NotImplementedError


class Citation(CustomDict):
    def __init__(self, cites, **kwargs):
        for cite in cites:
            cite.citation = self
        self.cites = cites
        super(Citation, self).__init__(kwargs)

    def __repr__(self):
        cites = ', '.join([cite.key for cite in self.cites])
        return '{}({})'.format(self.__class__.__name__, cites)


class CitationItem(CustomDict):
    def __init__(self, key, bibliography=None, **args):
        self.key = key.lower()
        optional = {'locator', 'prefix', 'suffix'}
        super(CitationItem, self).__init__(args, optional=optional)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.key)

    @property
    def bibliography(self):
        return self.citation.bibliography

    @property
    def reference(self):
        return self.bibliography.source[self.key]

    @property
    def number(self):
        return self.bibliography.keys.index(self.key) + 1

    @property
    def has_locator(self):
        return 'locator' in self

    def get_field(self, field):
        string = self.reference.get(field)
        if string is not None:
            return self.bibliography.formatter.preformat(string)

    def is_bad(self):
        return self.key not in self.bibliography.keys


class Locator(object):
    def __init__(self, label, identifier):
        self.label = label
        self.identifier = identifier


class BibliographySource(dict):
    def add(self, entry):
        self[entry.key] = entry


from . import bibtex, json



# coding: utf-8

from unittest import TestCase

from citeproc.source.bibtex import BibTeX
from citeproc.source.bibtex.bibtex import split_names, split_name, parse_name


class TestBibTeX(TestCase):
    def test_split_names(self):
        for names, parts in SPLIT_NAMES:
            print(names, split_names(names))
            self.assertEqual(split_names(names), parts)

    def test_split_name(self):
        for name, parts in SPLIT_NAME:
            self.assertEqual(split_name(name), parts)

    def test_parse_name(self):
        for name, reference in DECORET_NAMES:
            print('{:24}  {}'.format(name, parse_name(name)))
            self.assertEqual(parse_name(name), reference)
        for name, parts in PYBTEX_NAMES:
            reference = tuple(' '.join(pieces) if pieces else None
                              for pieces in parts)
            print('{:24}  {}'.format(name, parse_name(name)))
            self.assertEqual(parse_name(name), reference)
        for name, reference in EXTRA_NAMES:
            print('{:24}  {}'.format(name, parse_name(name)))
            self.assertEqual(parse_name(name), reference)

    def test_date_months(self):
        for january in ['jan', 'JAN', '01']:
            self.assertEqual(BibTeX._parse_month(january),
                             ({'month': 1}, ) * 2)
        self.assertEqual(BibTeX._parse_month("10~jan"),
                         ({'month': 1, 'day': 10}, ) * 2)
        self.assertEqual(BibTeX._parse_month("jul~4"),
                         ({'month': 7, 'day': 4}, ) * 2)

    def test_pages(self):
        def test(bibtex, csl):
            self.assertEqual(BibTeX._bibtex_to_csl_pages(bibtex), csl)

        test('313', '313')
        test('R112', 'R112')
        test('12+', '12')
        # page ranges
        test('1--3', '1-3')
        test('1-3', '1-3')
        test('i-iv', 'i-iv')
        test('11.1--11.27', '11.1-11.27')
        test('S43--S67', 'S43-S67')
        # Arabic letters
        test('ا-ي', 'ا-ي')


SPLIT_NAMES = [
    ('AA BB', ['AA BB']),
    ('AA and BB', ['AA', 'BB']),
    ('AA{ and }BB', ['AA{ and }BB']),
    ('AA{ }and BB', ['AA{ }and BB']),
]


SPLIT_NAME = [
    ('aa bb cc', [['aa', 'bb', 'cc']]),
    ('aa{ }bb cc', [['aa{ }bb', 'cc']]),
    ('aa{, }bb cc', [['aa{, }bb', 'cc']]),
    ('aa {bb} cc', [['aa', '{bb}', 'cc']]),
    ('aa bb, cc', [['aa', 'bb'], ['cc']]),
]


# http://maverick.inria.fr/~Xavier.Decoret/resources/xdkbibtex/bibtex_summary.html
DECORET_NAMES = [
    # name string          First        von           Last          Jr
    ( 'AA BB'          , ( 'AA'       , None        , 'BB'        , None     )),  # Testing simple case with no von.
    ( 'AA'             , (  None      , None        , 'AA'        , None     )),  # Testing that Last cannot be empty.
    ( 'AA bb'          , ( 'AA'       , None        , 'bb'        , None     )),  # Idem.
    ( 'aa'             , (  None      , None        , 'aa'        , None     )),  # Idem.
    ( 'AA bb CC'       , ( 'AA'       , 'bb'        , 'CC'        , None     )),  # Testing simple von.
    ( 'AA bb CC dd EE' , ( 'AA'       , 'bb CC dd'  , 'EE'        , None     )),  # Testing simple von (with inner uppercase words)
    ( 'AA 1B cc dd'    , ( 'AA 1B'    , 'cc'        , 'dd'        , None     )),  # Testing that digits are caseless (B fixes the case of 1B to uppercase).
    ( 'AA 1b cc dd'    , ( 'AA'       , '1b cc'     , 'dd'        , None     )),  # Testing that digits are caseless (b fixes the case of 1b to lowercase)
    ( 'AA {b}B cc dd'  , ( 'AA {b}B'  , 'cc'        , 'dd'        , None     )),  # Testing that pseudo letters are caseless.
    ( 'AA {b}b cc dd'  , ( 'AA'       , '{b}b cc'   , 'dd'        , None     )),  # Idem.
    ( 'AA {B}b cc dd'  , ( 'AA'       , '{B}b cc'   , 'dd'        , None     )),  # Idem.
    ( 'AA {B}B cc dd'  , ( 'AA {B}B'  , 'cc'        , 'dd'        , None     )),  # Idem.
    (r'AA \BB{b} cc dd', (r'AA \BB{b}', 'cc'        , 'dd'        , None     )),  # Testing that non letters are case less (in particular show how latex command are considered).
    (r'AA \bb{b} cc dd', ( 'AA'       ,r'\bb{b} cc' , 'dd'        , None     )),  # Idem.
    ( 'AA {bb} cc DD'  , ( 'AA {bb}'  , 'cc'        , 'DD'        , None     )),  # Testing that caseless words are grouped with First primilarily and then with Last.
    ( 'AA bb {cc} DD'  , ( 'AA'       , 'bb'        , '{cc} DD'   , None     )),  # Idem.
    ( 'AA {bb} CC'     , ( 'AA {bb}'  , None        , 'CC'        , None     )),  # Idem.

    ( 'bb CC, AA'      , ('AA'        , 'bb'        , 'CC'        , None     )),  # Simple case. Case do not matter for First.
    ( 'bb CC, aa'      , ('aa'        , 'bb'        , 'CC'        , None     )),  # Idem.
    ( 'bb CC dd EE, AA', ('AA'        , 'bb CC dd'  , 'EE'        , None     )),  # Testing simple von (with inner uppercase).
    ( 'bb, AA'         , ('AA'        , None        , 'bb'        , None     )),  # Testing that the Last part cannot be empty.
    ( 'BB,'            , ( None       , None        , 'BB'        , None     )),  # Testing that first can be empty after coma

    ( 'bb CC,XX, AA'   , ('AA'        , 'bb'        , 'CC'        , 'XX'     )),  # Simple Jr. Case do not matter for it.
    ( 'bb CC,xx, AA'   , ('AA'        , 'bb'        , 'CC'        , 'xx'     )),  # Idem.
    ( 'BB,, AA'        , ('AA'        , None        , 'BB'        , None     )),  # Testing that jr can be empty in between comas.

    (r"Paul \'Emile Victor"   , (r"Paul \'Emile"  , None        , "Victor", None)),
    (r"Paul {\'E}mile Victor" , (r"Paul {\'E}mile", None        , "Victor", None)),
    (r"Paul \'emile Victor"   , ( "Paul"          , r"\'emile"  , "Victor", None)),
    (r"Paul {\'e}mile Victor" , ( "Paul"          , r"{\'e}mile", "Victor", None)),

    (r"Victor, Paul \'Emile"  , (r"Paul \'Emile"  , None        , "Victor", None)),
    (r"Victor, Paul {\'E}mile", (r"Paul {\'E}mile", None        , "Victor", None)),
    (r"Victor, Paul \'emile"  , (r"Paul \'emile"  , None        , "Victor", None)),
    (r"Victor, Paul {\'e}mile", (r"Paul {\'e}mile", None        , "Victor", None)),

    ('Dominique Galouzeau de Villepin',
      ('Dominique Galouzeau', 'de'            , 'Villepin'             , None)),
    ('Dominique {G}alouzeau de Villepin',
      ('Dominique'          , '{G}alouzeau de', 'Villepin'             , None)),
    ('Galouzeau de Villepin, Dominique',
      ('Dominique'          , None            , 'Galouzeau de Villepin', None)),
]

# from Pybtex - /pybtex/tests/parse_name_test.py
PYBTEX_NAMES = [
    ('A. E.                   Siegman', (['A.', 'E.'], [], ['Siegman'], [])),
    ('A. G. W. Cameron', (['A.', 'G.', 'W.'], [], ['Cameron'], [])),
    ('A. Hoenig', (['A.'], [], ['Hoenig'], [])),
    ('A. J. Van Haagen', (['A.', 'J.', 'Van'], [], ['Haagen'], [])),
    ('A. S. Berdnikov', (['A.', 'S.'], [], ['Berdnikov'], [])),
    ('A. Trevorrow', (['A.'], [], ['Trevorrow'], [])),
    ('Adam H. Lewenberg', (['Adam', 'H.'], [], ['Lewenberg'], [])),
    ('Addison-Wesley Publishing Company',
    (['Addison-Wesley', 'Publishing'], [], ['Company'], [])),
    ('Advogato (Raph Levien)', (['Advogato', '(Raph'], [], ['Levien)'], [])),
    ('Andrea de Leeuw van Weenen',
    (['Andrea'], ['de', 'Leeuw', 'van'], ['Weenen'], [])),
    ('Andreas Geyer-Schulz', (['Andreas'], [], ['Geyer-Schulz'], [])),
    ("Andr{\\'e} Heck", (["Andr{\\'e}"], [], ['Heck'], [])),
    ('Anne Br{\\"u}ggemann-Klein', (['Anne'], [], ['Br{\\"u}ggemann-Klein'], [])),
    ('Anonymous', ([], [], ['Anonymous'], [])),
    ('B. Beeton', (['B.'], [], ['Beeton'], [])),
    ('B. Hamilton Kelly', (['B.', 'Hamilton'], [], ['Kelly'], [])),
    ('B. V. Venkata Krishna Sastry',
    (['B.', 'V.', 'Venkata', 'Krishna'], [], ['Sastry'], [])),
    ('Benedict L{\\o}fstedt', (['Benedict'], [], ['L{\\o}fstedt'], [])),
    ('Bogus{\\l}aw Jackowski', (['Bogus{\\l}aw'], [], ['Jackowski'], [])),
    ('Christina A. L.\\ Thiele',
    (['Christina', 'A.', 'L.\\'], [], ['Thiele'], [])),
    ("D. Men'shikov", (['D.'], [], ["Men'shikov"], [])),
    ("Darko \\v{Z}ubrini{\\'c}", (['Darko'], [], ["\\v{Z}ubrini{\\'c}"], [])),
    ("Dunja Mladeni{\\'c}", (['Dunja'], [], ["Mladeni{\\'c}"], [])),
    ('Edwin V. {Bell, II}', (['Edwin', 'V.'], [], ['{Bell, II}'], [])),
    ('Frank G. {Bennett, Jr.}', (['Frank', 'G.'], [], ['{Bennett, Jr.}'], [])),
    ("Fr{\\'e}d{\\'e}ric Boulanger",
    (["Fr{\\'e}d{\\'e}ric"], [], ['Boulanger'], [])),
    ('Ford, Jr., Henry', (['Henry'], [], ['Ford'], ['Jr.'])),
    ('mr Ford, Jr., Henry', (['Henry'], ['mr'], ['Ford'], ['Jr.'])),
    ('Fukui Rei', (['Fukui'], [], ['Rei'], [])),
    ('G. Gr{\\"a}tzer', (['G.'], [], ['Gr{\\"a}tzer'], [])),
    ('George Gr{\\"a}tzer', (['George'], [], ['Gr{\\"a}tzer'], [])),
    ('Georgia K. M. Tobin', (['Georgia', 'K.', 'M.'], [], ['Tobin'], [])),
    ('Gilbert van den Dobbelsteen',
    (['Gilbert'], ['van', 'den'], ['Dobbelsteen'], [])),
    ('Gy{\\"o}ngyi Bujdos{\\\'o}', (['Gy{\\"o}ngyi'], [], ["Bujdos{\\'o}"], [])),
    ('Helmut J{\\"u}rgensen', (['Helmut'], [], ['J{\\"u}rgensen'], [])),
    ('Herbert Vo{\\ss}', (['Herbert'], [], ['Vo{\\ss}'], [])),
    ("H{\\'a}n Th{\\^e}\\llap{\\raise 0.5ex\\hbox{\\'{\\relax}}}                  Th{\\'a}nh",
    (["H{\\'a}n", "Th{\\^e}\\llap{\\raise 0.5ex\\hbox{\\'{\\relax}}}"],
    [],
    ["Th{\\'a}nh"],
    [])),
    ("H{\\`a}n Th\\^e\\llap{\\raise0.5ex\\hbox{\\'{\\relax}}}                  Th{\\`a}nh",
    (['H{\\`a}n', "Th\\^e\\llap{\\raise0.5ex\\hbox{\\'{\\relax}}}"],
    [],
    ['Th{\\`a}nh'],
    [])),
    ("J. Vesel{\\'y}", (['J.'], [], ["Vesel{\\'y}"], [])),
    ("Javier Rodr\\'{\\i}guez Laguna",
    (['Javier', "Rodr\\'{\\i}guez"], [], ['Laguna'], [])),
    ("Ji\\v{r}\\'{\\i} Vesel{\\'y}",
    (["Ji\\v{r}\\'{\\i}"], [], ["Vesel{\\'y}"], [])),
    ("Ji\\v{r}\\'{\\i} Zlatu{\\v{s}}ka",
    (["Ji\\v{r}\\'{\\i}"], [], ['Zlatu{\\v{s}}ka'], [])),
    ("Ji\\v{r}{\\'\\i} Vesel{\\'y}",
    (["Ji\\v{r}{\\'\\i}"], [], ["Vesel{\\'y}"], [])),
    ("Ji\\v{r}{\\'{\\i}}Zlatu{\\v{s}}ka",
    ([], [], ["Ji\\v{r}{\\'{\\i}}Zlatu{\\v{s}}ka"], [])),
    ('Jim Hef{}feron', (['Jim'], [], ['Hef{}feron'], [])),
    ('J{\\"o}rg Knappen', (['J{\\"o}rg'], [], ['Knappen'], [])),
    ('J{\\"o}rgen L. Pind', (['J{\\"o}rgen', 'L.'], [], ['Pind'], [])),
    ("J{\\'e}r\\^ome Laurens", (["J{\\'e}r\\^ome"], [], ['Laurens'], [])),
    ('J{{\\"o}}rg Knappen', (['J{{\\"o}}rg'], [], ['Knappen'], [])),
    ('K. Anil Kumar', (['K.', 'Anil'], [], ['Kumar'], [])),
    ("Karel Hor{\\'a}k", (['Karel'], [], ["Hor{\\'a}k"], [])),
    ("Karel P\\'{\\i}{\\v{s}}ka", (['Karel'], [], ["P\\'{\\i}{\\v{s}}ka"], [])),
    ("Karel P{\\'\\i}{\\v{s}}ka", (['Karel'], [], ["P{\\'\\i}{\\v{s}}ka"], [])),
    ("Karel Skoup\\'{y}", (['Karel'], [], ["Skoup\\'{y}"], [])),
    ("Karel Skoup{\\'y}", (['Karel'], [], ["Skoup{\\'y}"], [])),
    ('Kent McPherson', (['Kent'], [], ['McPherson'], [])),
    ('Klaus H{\\"o}ppner', (['Klaus'], [], ['H{\\"o}ppner'], [])),
    ('Lars Hellstr{\\"o}m', (['Lars'], [], ['Hellstr{\\"o}m'], [])),
    ('Laura Elizabeth Jackson',
    (['Laura', 'Elizabeth'], [], ['Jackson'], [])),
    ("M. D{\\'{\\i}}az", (['M.'], [], ["D{\\'{\\i}}az"], [])),
    ('M/iche/al /O Searc/oid', (['M/iche/al', '/O'], [], ['Searc/oid'], [])),
    ("Marek Ry{\\'c}ko", (['Marek'], [], ["Ry{\\'c}ko"], [])),
    ('Marina Yu. Nikulina', (['Marina', 'Yu.'], [], ['Nikulina'], [])),
    ("Max D{\\'{\\i}}az", (['Max'], [], ["D{\\'{\\i}}az"], [])),
    ('Merry Obrecht Sawdey', (['Merry', 'Obrecht'], [], ['Sawdey'], [])),
    ("Miroslava Mis{\\'a}kov{\\'a}",
    (['Miroslava'], [], ["Mis{\\'a}kov{\\'a}"], [])),
    ('N. A. F. M. Poppelier', (['N.', 'A.', 'F.', 'M.'], [], ['Poppelier'], [])),
    ('Nico A. F. M. Poppelier',
    (['Nico', 'A.', 'F.', 'M.'], [], ['Poppelier'], [])),
    ('Onofrio de Bari', (['Onofrio'], ['de'], ['Bari'], [])),
    ("Pablo Rosell-Gonz{\\'a}lez", (['Pablo'], [], ["Rosell-Gonz{\\'a}lez"], [])),
    ('Paco La                  Bruna', (['Paco', 'La'], [], ['Bruna'], [])),
    ('Paul                  Franchi-Zannettacci',
    (['Paul'], [], ['Franchi-Zannettacci'], [])),
    ('Pavel \\v{S}eve\\v{c}ek', (['Pavel'], [], ['\\v{S}eve\\v{c}ek'], [])),
    ('Petr Ol{\\v{s}}ak', (['Petr'], [], ['Ol{\\v{s}}ak'], [])),
    ("Petr Ol{\\v{s}}{\\'a}k", (['Petr'], [], ["Ol{\\v{s}}{\\'a}k"], [])),
    ('Primo\\v{z} Peterlin', (['Primo\\v{z}'], [], ['Peterlin'], [])),
    ('Prof. Alban Grimm', (['Prof.', 'Alban'], [], ['Grimm'], [])),
    ("P{\\'e}ter Husz{\\'a}r", (["P{\\'e}ter"], [], ["Husz{\\'a}r"], [])),
    ("P{\\'e}ter Szab{\\'o}", (["P{\\'e}ter"], [], ["Szab{\\'o}"], [])),
    ('Rafa{\\l}\\.Zbikowski', ([], [], ['Rafa{\\l}\\.Zbikowski'], [])),
    ('Rainer Sch{\\"o}pf', (['Rainer'], [], ['Sch{\\"o}pf'], [])),
    ('T. L. (Frank) Pappas', (['T.', 'L.', '(Frank)'], [], ['Pappas'], [])),
    ('TUG 2004 conference', (['TUG', '2004'], [], ['conference'], [])),
    # ('TUG {\\sltt DVI} Driver Standards Committee',
    # (['TUG', '{\\sltt DVI}', 'Driver', 'Standards'], [], ['Committee'], [])),
    ('University of M{\\"u}nster',
    (['University'], ['of'], ['M{\\"u}nster'], [])),
    ('Walter van der Laan', (['Walter'], ['van', 'der'], ['Laan'], [])),
    ('Wendy G.                  McKay', (['Wendy', 'G.'], [], ['McKay'], [])),
    ('Wendy McKay', (['Wendy'], [], ['McKay'], [])),
    ('W{\\l}odek Bzyl', (['W{\\l}odek'], [], ['Bzyl'], [])),
    ('\\LaTeX Project Team', (['\\LaTeX', 'Project'], [], ['Team'], [])),
    ('\\rlap{Lutz Birkhahn}', ([], [], ['\\rlap{Lutz Birkhahn}'], [])),
    ('{Jim Hef{}feron}', ([], [], ['{Jim Hef{}feron}'], [])),
    ('{Kristoffer H\\o{}gsbro Rose}',
    ([], [], ['{Kristoffer H\\o{}gsbro Rose}'], [])),
    ('{TUG} {Working} {Group} on a {\\TeX} {Directory}                  {Structure}',
    (['{TUG}', '{Working}', '{Group}'],
    ['on', 'a'],
    ['{\\TeX}', '{Directory}', '{Structure}'],
    [])),
    ('{The \\TUB{} Team}', ([], [], ['{The \\TUB{} Team}'], [])),
    ('{\\LaTeX} project team', (['{\\LaTeX}'], ['project'], ['team'], [])),
    ('{\\NTG{} \\TeX{} future working group}',
    ([], [], ['{\\NTG{} \\TeX{} future working group}'], [])),
    ('{{\\LaTeX\\,3} Project Team}',
    ([], [], ['{{\\LaTeX\\,3} Project Team}'], [])),
    ('Johansen Kyle, Derik Mamania M.',
    (['Derik', 'Mamania', 'M.'], [], ['Johansen', 'Kyle'], [])),
]


EXTRA_NAMES = [
    ('others', (None, None, 'others', None)),
    ('Charles Louis Xavier Joseph de la Vall{\’e}e Poussin',
      ('Charles Louis Xavier Joseph', 'de la', 'Vall{\’e}e Poussin', None)),
    # the following name of the Pybtex tests doesn't agree with BibTeX's output
    # (verified using Xavier Décoret's names.bst)
    ('TUG {\\sltt DVI} Driver Standards Committee',
      ('TUG', '{\\sltt DVI}', 'Driver Standards Committee', None)),
]




import os
import json

from . import types, formatter


DATA_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')

SCHEMA_PATH = os.path.join(DATA_PATH, 'schema', 'csl.rng')
LOCALES_PATH = os.path.join(DATA_PATH, 'locales')
STYLES_PATH = os.path.join(DATA_PATH, 'styles')


NAMES = ['author', 'collection_editor', 'composer', 'container_author',
         'editor', 'editorial_director', 'illustrator', 'interviewer',
         'original_author', 'recipient', 'translator']

DATES = ['accessed', 'container', 'event_date', 'issued', 'original_date',
         'submitted']

NUMBERS = ['chapter_number', 'collection_number', 'edition', 'issue', 'number',
           'number_of_pages', 'number_of_volumes', 'volume']

VARIABLES = (['abstract', 'annote', 'archive', 'archive_location',
              'archive_place', 'authority', 'call_number', 'citation_label',
              'citation_number', 'collection_title', 'container_title',
              'container_title_short', 'dimensions', 'DOI', 'event',
              'event_place', 'first_reference_note_number', 'genre', 'ISBN',
              'ISSN', 'jurisdiction', 'keyword', 'language', 'locator',
              'medium', 'note', 'original_publisher',
              'original_publisher_place', 'original_title', 'page',
              'page_first', 'PMCID', 'PMID', 'publisher', 'publisher_place',
              'references', 'section', 'source', 'status', 'title',
              'title_short', 'URL', 'version', 'year_suffix'] +
             NAMES + DATES + NUMBERS)

with open(os.path.join(LOCALES_PATH, 'locales.json'),
          encoding='utf-8') as file:
    locales_json = json.load(file)
    PRIMARY_DIALECTS = locales_json['primary-dialects']
    LANGUAGE_NAMES = locales_json['language-names']


from .frontend import CitationStylesStyle, CitationStylesBibliography
from .source import Citation, CitationItem, Locator




import re
import unicodedata
import os

from functools import cmp_to_key
from glob import glob
from operator import itemgetter

from lxml import etree

from . import NAMES, DATES, NUMBERS, PRIMARY_DIALECTS, LANGUAGE_NAMES
from .source import VariableError, DateRange, LiteralDate
from .string import String, join


# Base class

class SomewhatObjectifiedElement(etree.ElementBase):
    nsmap = {'cs': 'http://purl.org/net/xbiblio/csl',
             'xml': 'http://www.w3.org/XML/1998/namespace'}

    # TODO: what about multiple instances of the same name?
    def __getattr__(self, name):
        return self.find('cs:' + name, self.nsmap)


class CitationStylesElement(SomewhatObjectifiedElement):
    _default_options = {# global options
                        'initialize-with-hyphen': 'true',
                        'page-range-format': None,
                        'demote-non-dropping-particle': 'display-and-sort',

                        # inheritable name(s) options
                        'and': None,
                        'delimiter-precedes-et-al': 'contextual',
                        'delimiter-precedes-last': 'contextual',
                        'et-al-min': 0,
                        'et-al-use-first': 1,
                        'et-al-subsequent-min': 0,
                        'et-al-subsequent-use-first': 1,
                        'et-al-use-last': 'false',
                        'initialize-with': None,
                        'name-as-sort-order': None,
                        'sort-separator': ', ',

                        'name-form': 'long',
                        'name-delimiter': ', ',
                        'names-delimiter': ''}

    def get_root(self):
        return self.getroottree().getroot()

    def xpath_search(self, expression):
        return self.xpath(expression, namespaces=self.nsmap)

    @property
    def loc(self):
        full_xpath = self.getroottree().getpath(self)
        xpath = ''
        tree = []
        for i, node in enumerate(full_xpath.split('/')[1:]):
            xpath += '/' + node
            element = self.xpath(xpath)[0]
            namespace, tag = element.tag.split('}', 1)
            attribs = ''.join(' {}="{}"'.format(key, value)
                               for key, value in element.attrib.items())
            tree.append('{:>4}: {}<{}{}>'.format(element.sourceline,
                                                 i * '  ', tag, attribs))
        print('\n'.join(tree))

    def get_option(self, name):
        return self.get(name, self._default_options[name])

    def get_macro(self, name):
        expression = "cs:macro[@name='{}'][1]".format(name)
        return self.get_root().xpath_search(expression)[0]

    def get_layout(self):
        return self.xpath_search('./ancestor-or-self::cs:layout[1]')[0]

    def get_formatter(self):
        if isinstance(self.get_root(), Locale):
            return self.get_root().style.formatter
        else:
            return self.get_root().formatter

    def preformat(self, text):
        return self.get_formatter().preformat(text)

    def unicode_character(self, name):
        return self.preformat(unicodedata.lookup(name))

    def render(self, *args, **kwargs):
        return self.markup(self.process(*args, **kwargs))

    # TODO: Locale methods
    def get_term(self, name, form=None):
        if isinstance(self.get_root(), Locale):
            return self.get_root().get_term(name, form)
        else:
            for locale in self.get_root().locales:
                try:
                    return locale.get_term(name, form)
                except IndexError: # TODO: create custom exception
                    continue

    def get_date(self, form):
        for locale in self.get_root().locales:
            try:
                return locale.get_date(form)
            except IndexError:
                continue

    def get_locale_option(self, name):
        for locale in self.get_root().locales:
            try:
                return locale.get_option(name)
            except IndexError:
                continue


# Top level elements

class Style(CitationStylesElement):
    def set_locale_list(self, output_locale, validate=True):
        """Set up list of locales in which to search for localizable units"""
        from .frontend import CitationStylesLocale

        self.locales = []
        system_locales_added = set()

        def add_instyle_locale(locale):
            expr = ('./cs:locale[@xml:lang="{}"]'.format(locale)
                    if locale else './cs:locale[not(@xml:lang)]')
            results = self.xpath_search(expr)
            if results:
                locale_element, = results
                self.locales.append(locale_element)

        def add_system_locale(locale):
            if locale not in system_locales_added:
                csl_locale = CitationStylesLocale(locale, validate=validate)
                self.locales.append(csl_locale.root)
                system_locales_added.add(locale)

        # 1) (in-style locale) chosen dialect
        add_instyle_locale(output_locale)
        # 2) (in-style locale) matching language
        language = output_locale.split('-')[0]
        add_instyle_locale(language)
        # 3) (in-style locale) no language set
        add_instyle_locale(None)
        # 4) (locale files) chosen or primary dialect
        if output_locale in LANGUAGE_NAMES:
            add_system_locale(output_locale)
        # 5) (locale files) fall back to primary language dialect
        if language in PRIMARY_DIALECTS:
            add_system_locale(PRIMARY_DIALECTS[language])
        # 6) (locale files) default locale (en-US)
        add_system_locale('en-US')
        for locale in self.locales:
            locale.style = self


class Locale(CitationStylesElement):
    _default_options = {'limit-day-ordinals-to-day-1': 'false',
                        'punctuation-in-quote': 'false'}

    def get_term(self, name, form=None):
        attributes = "@name='{}'".format(name)
        if form is not None:
            attributes += " and @form='{}'".format(form)
        else:
            attributes += " and not(@form)"
        expr = './cs:term[{}]'.format(attributes)
        try:
            return self.terms.xpath_search(expr)[0]
        except AttributeError:
            raise IndexError

    def get_date(self, form):
        expr = "./cs:date[@form='{}']".format(form)
        return self.xpath_search(expr)[0]

    def get_option(self, name):
        options = self.find('cs:style-options', self.nsmap)
        if options is None:
            raise IndexError
        return options.get(name, self._default_options[name])

    def get_formatter(self):
        return self.style.formatter


class FormattingInstructions(object):
    def get_option(self, name):
        if name in self._default_options:
            return self.get(name, self._default_options[name])
        else:
            return self.get(name, self.get_root().get_option(name))

    def render(self, reference):
        raise NotImplementedError


class Citation(FormattingInstructions, CitationStylesElement):
    _default_options = {# disambiguation
                        'disambiguate-add-names': False,
                        'disambiguate-add-givenname': False,
                        'givenname-disambiguation-rule': 'all-names',
                        'disambiguate-add-year-suffix': False,

                        # citation collapsing
                        'collapse': None,
                        'year-suffix-delimiter': None,
                        'after-collapse-delimiter': None,

                        # note distance
                        'near-note-distance': 5}

    def render(self, citation, cites, callback):
        self.cites = cites
        return self.layout.render_citation(citation, callback)


class Bibliography(FormattingInstructions, CitationStylesElement):
    _default_options = {# whitespace
                        'hanging-indent': False,
                        'second-field-align': None,
                        'line-spacing': 1,
                        'entry-spacing': 1,

                        # reference grouping
                        'subsequent-author-substitute': None}

    def sort(self, citation_items):
        return self.layout.sort_bibliography(citation_items)

    def render(self, citation_items):
        return self.layout.render_bibliography(citation_items)


# Style behavior

class Formatted(object):
    def format(self, string):
        if isinstance(string, (int, float)):
            string = str(string)
        text = self.font_style(string)
        text = self.font_variant(text)
        text = self.font_weight(text)
        text = self.text_decoration(text)
        text = self.vertical_align(text)
        return text

    def font_style(self, text):
        formatter = self.get_formatter()
        font_style = self.get('font-style', 'normal')
        if font_style == 'normal':
            formatted = text
        elif font_style == 'italic':
            formatted = formatter.Italic(text)
        elif font_style == 'oblique':
            formatted = formatter.Oblique(text)
        return formatted

    def font_variant(self, text):
        formatter = self.get_formatter()
        font_variant = self.get('font-variant', 'normal')
        if font_variant == 'normal':
            formatted = text
        elif font_variant == 'small-caps':
            formatted = formatter.SmallCaps(text)
        return formatted

    def font_weight(self, text):
        formatter = self.get_formatter()
        font_weight = self.get('font-weight', 'normal')
        if font_weight == 'normal':
            formatted = text
        elif font_weight == 'bold':
            formatted = formatter.Bold(text)
        elif font_weight == 'light':
            formatted = formatter.Light(text)
        return formatted

    def text_decoration(self, text):
        formatter = self.get_formatter()
        text_decoration = self.get('text-decoration', 'none')
        if text_decoration == 'none':
            formatted = text
        elif text_decoration == 'underline':
            formatted = formatter.Underline(text)
        return formatted

    def vertical_align(self, text):
        formatter = self.get_formatter()
        vertical_align = self.get('vertical-align', 'baseline')
        if vertical_align == 'baseline':
            formatted = text
        elif vertical_align == 'sup':
            formatted = formatter.Superscript(text)
        elif vertical_align == 'sub':
            formatted = formatter.Subscript(text)
        return formatted


class Affixed(object):
    def wrap(self, string):
        if string is not None:
            prefix = self.get('prefix', '')
            suffix = self.get('suffix', '')
            return prefix + string + suffix
        return None


class Delimited(object):
    def join(self, strings, default_delimiter=''):
        delimiter = self.get('delimiter', default_delimiter)
        try:
            text = join((s for s in strings if s is not None), delimiter)
        except:
            text = String('')
        return text


class Displayed(object):
    pass


class Quoted(object):
    def quote(self, string):
        piq = self.get_locale_option('punctuation-in-quote').lower() == 'true'
        if self.get('quotes', 'false').lower() == 'true':
            open_quote = self.get_term('open-quote').single
            close_quote = self.get_term('close-quote').single
            string = open_quote + string + close_quote
##            quoted_string = QuotedString(string, open_quote, close_quote, piq)
        return string


class StrippedPeriods(object):
    def strip_periods(self, string):
        strip_periods = self.get('strip-periods', 'false').lower() == 'true'
        if strip_periods:
            string = string.replace('.', '')
        return string


class TextCased(object):
    _stop_words = ['a', 'an', 'and', 'as', 'at', 'but', 'by', 'down', 'for',
                   'from', 'in', 'into', 'nor', 'of', 'on', 'onto', 'or',
                   'over', 'so', 'the', 'till', 'to', 'up', 'via', 'with',
                   'yet']

    def case(self, text, language=None):
        text_case = self.get('text-case')
        if text_case is not None:
            if language != 'en' and text_case == 'title':
                text_case = 'sentence'
            if text_case == 'lowercase':
                text = text.lower()
            elif text_case == 'uppercase':
                text = text.upper()
            elif text_case == 'capitalize-first':
                text = text.capitalize_first()
            elif text_case == 'capitalize-all':
                output = []
                for word in text.words():
                    word = word.capitalize_first()
                    output.append(word)
                text = ' '.join(output)
            elif text_case == 'title':
                output = []
                prev = ':'
                for word in text.words():
                    if not text.isupper() and not word.isupper():
                        word = word.soft_lower()
                        if (str(word) not in self._stop_words or
                            prev in (':', '.')):
                            word = word.capitalize_first()
                    prev = word[-1]
                    output.append(word)
                text = ' '.join(output)
            elif text_case == 'sentence':
                output = []
                for i, word in enumerate(text.words()):
                    if not text.isupper() and not word.isupper():
                        word = word.soft_lower()
                    if i == 0:
                        word = word.capitalize_first()
                    output.append(word)
                text = ' '.join(output)
        return text


# Locale elements

class Term(CitationStylesElement):
    @property
    def single(self):
        try:
            text = self.find('cs:single', self.nsmap).text
        except AttributeError:
            text = self.text
        text = self.preformat(text or '')
        return String(text)

    @property
    def multiple(self):
        try:
            text = self.find('cs:multiple', self.nsmap).text
        except AttributeError:
            text = self.text
        text = self.preformat(text or '')
        return String(text)


# Sorting elements

class Sort(CitationStylesElement):
    def sort(self, items, context):
        # custom sort function to push items with None keys to bottom
        def multi_key_sort(items, keys, descending):
            lst = zip(items, *keys)
            comparers = [(itemgetter(i + 1), descending[i])
                         for i in range(len(keys))]
            def mycmp(left, right):
                for getter, desc in comparers:
                    left_key, right_key = getter(left), getter(right)
                    if left_key is not None and right_key is not None:
                        try:
                            left_key = str(left_key.lower())
                            right_key = str(right_key.lower())
                        except AttributeError:
                            pass
                        try:
                            left_key, right_key = (int(str(left_key)),
                                                   int(str(right_key)))
                        except ValueError:
                            pass
                        result = (left_key > right_key) - (left_key < right_key)
                        if result:
                            return -1 * result if desc else result
                    elif left_key is not None:
                        return -1
                    elif right_key is not None:
                        return 1
                    else:
                        continue
                else:
                    return 0

            sorted_lst = sorted(lst, key=cmp_to_key(mycmp))
            return [item[0] for item in sorted_lst]

        sort_descending = []
        sort_keys = []
        for key in self.findall('cs:key', self.nsmap):
            descending = key.get('sort', 'ascending').lower() == 'descending'
            sort_descending.append(descending)
            sort_keys.append(key.sort_keys(items, context))

        return multi_key_sort(items, sort_keys, sort_descending)


class Key(CitationStylesElement):
    def sort_keys(self, items, context):
        if 'variable' in self.attrib:
            variable = self.get('variable').replace('-', '_')
            if variable in NAMES:
                sort_keys = [self._format_name(item, variable)
                             for item in items]
            elif variable in DATES:
                sort_keys = []
                for item in items:
                    date = item.reference.get(variable)
                    if date is not None:
                        sort_keys.append(date.sort_key())
                    else:
                        sort_keys.append(None)
            elif variable in NUMBERS:
                sort_keys = [self._format_number(item, variable)
                             for item in items]
            elif variable == 'citation_number':
                sort_keys = [item.number for item in items]
            else:
                sort_keys = [item.get_field(variable) for item in items]
        elif 'macro' in self.attrib:
            layout = context.get_layout()
            # override name options
            sort_options = {'name-as-sort-order': 'all'}
            for option in ('names-min', 'names-use-first', 'names-use-last'):
                if option in self.attrib:
                    name = option.replace('names', 'et-al')
                    sort_options[name] = self.get(option)
            macro = self.get_macro(self.get('macro'))
            sort_keys = []
            for item in items:
                layout.repressed = {}
                sort_key = macro.render(item, context=context,
                                        sort_options=sort_options)
                sort_keys.append(sort_key)

        return sort_keys

    def _format_name(self, item, variable):
        names = item.reference.get(variable)
        if names is not None:
            output = []
            for name in names:
                demote_ndp = self.get_root().get('demote-non-dropping-particle',
                                                 'display-and-sort').lower()
                sort_separator = self._default_options['sort-separator']

                # TODO: encapsulate in function (to share with Name)
                given, family, dp, ndp, suffix = name.parts()
                if demote_ndp in ('sort-only', 'display-and-sort'):
                    given = ' '.join([n for n in (given, dp, ndp) if n])
                else:
                    family = ' '.join([n for n in (ndp, family) if n])
                    given = ' '.join([n for n in (given, dp) if n])

                order = family, given, suffix
                output.append(sort_separator.join([n for n in order if n]))
            return ';'.join(output)
        else:
            return None

    def _format_number(self, item, variable):
        date = item.reference.get(variable)
        if date is not None:
            try:
                return str(Number.re_numeric.match(date).group(1))
            except AttributeError:
                return date
        else:
            return None


# Rendering elements

class Parent(object):
    def calls_variable(self):
        return any([child.calls_variable() for child in self.getchildren()])

    def process_children(self, item, **kwargs):
        output = []
        for child in self.iterchildren():
            try:
                text = child.process(item, **kwargs)
                if text is not None:
                    output.append(text)
            except VariableError:
                pass
        if output:
            return ''.join(output)
        else:
            return None

    def render_children(self, item, **kwargs):
        output = []
        for child in self.iterchildren():
            try:
                text = child.render(item, **kwargs)
                if text is not None:
                    output.append(text)
            except VariableError:
                pass
        if output:
            return join(output)
        else:
            return None


class Macro(CitationStylesElement, Parent):
    def process(self, item, context=None, sort_options=None):
        return self.process_children(item, context=context,
                                    sort_options=sort_options)

    def render(self, item, context=None, sort_options=None):
        return self.render_children(item, context=context,
                                    sort_options=sort_options)


class Layout(CitationStylesElement, Parent, Formatted, Affixed, Delimited):
    def render_citation(self, citation, callback):
        # first sort citation items according to bibliography order
        bibliography = citation.bibliography
        good_cites = [cite for cite in citation.cites if not cite.is_bad()]
        bad_cites = [cite for cite in citation.cites if cite.is_bad()]
        good_cites.sort(key=lambda item: bibliography.keys.index(item.key))
        # sort using citation/sort element
        if self.getparent().sort is not None:
            good_cites = self.getparent().sort.sort(good_cites, self)
        out = []
        for item in good_cites:
            self.repressed = {}
            prefix = item.get('prefix', '')
            suffix = item.get('suffix', '')
            try:
                output = self.render_children(item)
                if output is not None:
                    text = prefix + output + suffix
                    out.append(text)
                    self.getparent().cites.append(item)
            except VariableError:
                pass
        for item in bad_cites:
            callback_value = callback(item)
            out.append(callback_value or '{}?'.format(item.key))
        return self.format(self.wrap(self.join(out)))

    def sort_bibliography(self, citation_items):
        sort = self.getparent().find('cs:sort', self.nsmap)
        if sort is not None:
            citation_items = sort.sort(citation_items, self)
        return citation_items

    def render_bibliography(self, citation_items):
        output_items = []
        for item in citation_items:
            self.repressed = {}
            text = self.format(self.wrap(self.render_children(item)))
            if text is not None:
                output_items.append(text)
        return output_items


class FormatNumber(object):
    def _process(self, value, variable):
        page_range_delimiter = (self.get_term('page-range-delimiter').single
                                if variable.startswith('page') else None)
        range_delimiter = (page_range_delimiter
                           or self.unicode_character('EN DASH'))

        en_dash = unicodedata.lookup('EN DASH')
        def format_number_or_range(item):
            try:
                first, last = (number.strip() for number
                               in item.replace(en_dash, '-').split('-'))
            except ValueError:
                return self.format_number(item.strip())
            first = self.format_number(first)
            if variable == 'page-first':
                return first
            last = self.format_number(self._format_last_page(first, last)
                                      if variable == 'page' else last)
            return join((first, last), range_delimiter)

        amp_delimiter = ' ' + self.unicode_character('AMPERSAND') + ' '
        return join((join((format_number_or_range(item)
                           for item in comma_item.split('&')), amp_delimiter)
                     for comma_item in value.split(',')), delimiter=', ')

    def _format_last_page(self, first, last):
        def find_common(first, last):
            count = 0
            for count, (f, l) in enumerate(zip(first, last)):
                if f != l:
                    return count
            return count + 1

        range_format = self.get_root().get_option('page-range-format')
        common = find_common(first, last)
        if range_format == 'chicago':
            m = re.search(r'\d+', first)
            first_number = int(m.group())
            if first_number < 100 or first_number % 100 == 0:
                range_format = 'expanded'
            elif len(first) >= 4 and common < 2:
                range_format = 'expanded'
            elif first_number % 100 in range(1, 10):
                range_format = 'minimal'
            elif first_number % 100 in range(10, 100):
                range_format = 'minimal-two'
        if range_format in ('expanded', None):
            index = 0
        elif range_format == 'minimal':
            index = common
        elif range_format == 'minimal-two':
            index = min(common, len(first) - 2)
        return last[index:]

    def format_number(self, number):
        return str(number)


class Text(CitationStylesElement, FormatNumber, Formatted, Affixed, Quoted,
           TextCased, StrippedPeriods):
    generated_variables = ('year-suffix', 'citation-number')

    def calls_variable(self):
        if 'variable' in self.attrib:
            return self.get('variable') not in self.generated_variables
        elif 'macro' in self.attrib:
            return self.get_macro(self.get('macro')).calls_variable()
        else:
            return False

    def render(self, *args, **kwargs):
        text, language = self.process(*args, **kwargs)
        return self.markup(text, language)

    def process(self, item, context=None, **kwargs):
        if context is None:
            context = self

        try:
            language = item.reference.language[:2]
        except VariableError:
            language = self.get_root().get('default-locale', 'en')[:2]
        if 'variable' in self.attrib:
            text = self._variable(item, context)
        elif 'macro' in self.attrib:
            text = self.get_macro(self.get('macro')).render(item, context)
        elif 'term' in self.attrib:
            text = self._term(item)
        elif 'value' in self.attrib:
            text = String(self.preformat(self.get('value')))

        return text, language

    def _variable(self, item, context):
        variable = self.get('variable')
        repressed = context.get_layout().repressed
        if self.tag in repressed and variable in repressed[self.tag]:
            return None

        if self.get('form') == 'short':
            short_variable = variable + '-short'
            if short_variable.replace('-', '_') in item.reference:
                variable = short_variable

        if variable.startswith('page'):
            text = self._process(item.reference.page, variable)
        elif variable == 'citation-number':
            text = item.number
        elif variable == 'locator':
            en_dash = self.unicode_character('EN DASH')
            text = str(item.locator.identifier).replace('-', en_dash)
        else:
            text = item.reference[variable.replace('-', '_')]

        return text

    def _term(self, item):
        form = self.get('form', 'long')
        plural = self.get('plural', 'false').lower() == 'true'
        if form == 'long':
            form = None
        term = self.get_term(self.get('term'), form)

        if plural:
            text = term.multiple
        else:
            text = term.single

        return text

    def markup(self, text, language):
        if text:
            tmp = self.format(self.case(self.strip_periods(text), language))
            return self.wrap(self.quote(tmp))
        else:
            return None


class Date(CitationStylesElement, Parent, Formatted, Affixed, Delimited):
    def calls_variable(self):
        return True

    def is_locale_date(self):
        expr = './ancestor::cs:locale[1]'
        try:
            self.xpath_search(expr)[0]
            return True
        except IndexError:
            return False

    def render_single_date(self, date, show_parts=None, context=None):
        form = self.get('form')
        if context != self:
            parts = self.parts(date, show_parts, context)
        else:
            parts = self.parts(date, show_parts)
        if parts:
            style_context = context if self.is_locale_date() else self
            return style_context.join(parts)
        else:
            return None

    def render_date_range(self, date_range, show_parts=None, context=None):
        same_show_parts = []
        if date_range.end.is_nil():
            same = None
            diff_begin = self.render_single_date(date_range.begin, show_parts,
                                                 context)
            diff_end = ''
        else:
            if date_range.begin.year == date_range.end.year:
                show_parts.remove('year')
                same_show_parts.append('year')
                try:
                    if ('month' in show_parts and
                        date_range.begin.month == date_range.end.month):
                        show_parts.remove('month')
                        same_show_parts.append('month')
                        try:
                            if ('day' in show_parts and
                                date_range.begin.day == date_range.end.day):
                                show_parts.remove('day')
                                same_show_parts.append('day')
                        except AttributeError:
                            show_parts.remove('day')
                except AttributeError:
                    show_parts.remove('month')

            same = self.render_single_date(date_range.end, same_show_parts,
                                           context)
            diff_begin = self.render_single_date(date_range.begin, show_parts,
                                                 context)
            diff_end = self.render_single_date(date_range.end, show_parts,
                                               context)

        if not (diff_begin and diff_begin):
            return None

        diff = (diff_begin.rstrip() + self.unicode_character('EN DASH') +
                diff_end)
        if same:
            text = context.join([diff, same.rstrip()])
        else:
            text = diff

        return text

    def process(self, item, variable=None, show_parts=None, context=None,
                **kwargs):
        if variable is None:
            variable = self.get('variable')
        if show_parts is None:
            show_parts = ['year', 'month', 'day']
        if context is None:
            context = self

        form = self.get('form')
        date_parts = self.get('date-parts')
        if not self.is_locale_date() and form is not None:
            localized_date = self.get_date(form)
            if date_parts is not None:
                show_parts = date_parts.split('-')
            return localized_date.render(item, variable,
                                         show_parts=show_parts, context=self)
        else:
            date_or_range = item.reference[variable.replace('-', '_')]
            if not date_or_range:
                text = None
            elif isinstance(date_or_range, LiteralDate):
                text = date_or_range.text
            elif isinstance(date_or_range, DateRange):
                text = self.render_date_range(date_or_range, show_parts,
                                              context)
            else:
                text = self.render_single_date(date_or_range, show_parts,
                                               context)
            if text is not None:
                style_context = context if self.is_locale_date() else self
                return style_context.wrap(text)
            else:
                return None

    def parts(self, date, show_parts, context=None):
        output = []
        for part in self.iterchildren():
            if part.get('name') in show_parts:
                try:
                    part_text = part.render(date, context)
                    if part_text is not None:
                        output.append(part_text)
                except VariableError:
                    pass
        return output

    def markup(self, text):
        # TODO: fix
        return text


class Date_Part(CitationStylesElement, Formatted, Affixed, TextCased,
                StrippedPeriods):
    def process(self, date, context=None):
        name = self.get('name')
        range_delimiter = self.get('range-delimiter', '-')
        attrib = self.attrib

        if context is None:
            context = self
        try:
            expr = './cs:date-part[@name="{}"]'.format(name)
            attrib.update(dict(context.xpath_search(expr)[0].attrib))
        except (AttributeError, IndexError):
            pass

        if name == 'day':
            form = self.get('form', 'numeric')
            if (form == 'ordinal'
                and self.get_locale_option('limit-day-ordinals-to-day-1')
                    .lower() == 'true'
                and date.day > 1):
                form = 'numeric'

            if form == 'numeric':
                text = date.day
            elif form == 'numeric-leading-zeros':
                text = '{:02}'.format(date.day)
            elif form == 'ordinal':
                text = to_ordinal(date.day, context)
        elif name == 'month':
            form = self.get('form', 'long')
            strip_periods = self.get('form', False)
            try:
                index = date.month
                term = 'month'
            except VariableError:
                index = date.season
                term = 'season'

            if form == 'long':
                text = context.get_term('{}-{:02}'.format(term, index)).single
            elif form == 'short':
                term = context.get_term('{}-{:02}'.format(term, index), 'short')
                text = term.single
            else:
                assert term == 'month'
                if form == 'numeric':
                    text = '{}'.format(index)
                elif form == 'numeric-leading-zeros':
                    text = '{:02}'.format(index)
        elif name == 'year':
            form = self.get('form', 'long')
            if form == 'long':
                text = str(abs(date.year))
                if date.year < 0:
                    text += context.get_term('bc').single
                elif date.year < 1000:
                    text += context.get_term('ad').single
            elif form == 'short':
                text = str(date.year)[-2:]

        return text

    def markup(self, text):
        if text:
            return self.wrap(self.format(self.case(self.strip_periods(text))))
        else:
            return None


class Number(CitationStylesElement, FormatNumber, Formatted, Affixed, Displayed,
             TextCased, StrippedPeriods):
    def calls_variable(self):
        return True

    def process(self, item, context=None, **kwargs):
        variable = self.get('variable')
        if variable == 'locator':
            try:
                variable = item.locator.label
                value = item.locator.identifier
            except KeyError:
                return None
        elif variable == 'page-first':
            value = item.reference.page
        else:
            value = item.reference[variable]
        return self._process(value, variable)

    def format_number(self, number):
        form = self.get('form', 'numeric')
        try:
            number = int(number)
        except ValueError:
            return number
        if form == 'numeric':
            text = str(number)
        elif form == 'ordinal' or form == 'long-ordinal' and number > 10:
            text = to_ordinal(number, self)
        elif form == 'long-ordinal':
            text = self.get_term('long-ordinal-{:02}'.format(number)).single
        elif form == 'roman':
            text = romanize(number).lower()
        return text

    def markup(self, text):
        if text:
            return self.wrap(self.format(self.case(self.strip_periods(text))))
        else:
            return None


class Names(CitationStylesElement, Parent, Formatted, Affixed, Delimited):
    def calls_variable(self):
        return True

    def get_parent_delimiter(self, context=None):
        expr = './ancestor::*[self::cs:citation or self::cs:bibliography][1]'
        if context is None:
            context = self
        parent = context.xpath_search(expr)[0]
        return parent.get_option('names-delimiter')

    def substitute(self):
        expr = './cs:substitute[1]'
        try:
            result = self.xpath_search(expr)[0]
        except IndexError:
            result = None
        return result

    def process(self, item, names_context=None, context=None, **kwargs):
        if context is None:
            context = self
        if names_context is None:
            names_context = self

        roles = self.get('variable').split()
        try:
            ed_trans = (set(roles) == set(['editor', 'translator']) and
                        item.reference.editor == item.reference.translator and
                        self.get_term('editortranslator').getchildren())
            if ed_trans:
                roles = ['editor']
        except VariableError:
            ed_trans = False

        output = []
        for role in roles:
            if role in item.reference:
                name_elem = names_context.name
                if name_elem is None:
                    name_elem = Name()
                    names_context.insert(0, name_elem)
                text = name_elem.render(item, role, context=context, **kwargs)
                plural = len(item.reference[role]) > 1
                try:
                    if ed_trans:
                        role = 'editortranslator'
                    label_element = names_context.label
                    label = label_element.render(item, role, plural, **kwargs)
                    if label is not None:
                        if label_element is names_context.getchildren()[0]:
                            text = label + text
                        else:
                            text = text + label
                except AttributeError:
                    pass
                output.append(text)

        if output:
            try:
                total = sum(output)
            except TypeError:
                is_int = False
            else:
                is_int = isinstance(total, int)
            if is_int:
                text = str(total) if total > 0 else None
            else:
                text = self.join(output, self.get_parent_delimiter(context))
        else:
            substitute = self.substitute()
            if substitute is not None:
                text = substitute.render(item, context=context, **kwargs)

        try:
            return text
        except NameError:
            raise VariableError

    def markup(self, text):
        if text:
            return self.wrap(self.format(text))
        else:
            return None


class Name(CitationStylesElement, Formatted, Affixed, Delimited):
    def get_option(self, name, context=None, sort_options=None):
        try:
            value = sort_options[name]
        except (TypeError, KeyError):
            expr = ('./ancestor::*[self::cs:citation or '
                                  'self::cs:bibliography][1]')
            if context is None:
                context = self
            parent = context.xpath_search(expr)[0]
            if name in ('form', 'delimiter'):
                value = self.get(name, parent.get_option('name-' + name))
            else:
                value = self.get(name, parent.get_option(name))

        if name in ('initialize-with-hyphen', 'et-al-use-last'):
            value = value.lower() == 'true'
        elif name.startswith('et-al'):
            value = int(value)

        return value

    def et_al(self):
        expr = './following-sibling::cs:et-al[1]'
        try:
            result = self.xpath_search(expr)[0].render()
        except IndexError:
            result = self.get_term('et-al').single
        return result

    def process(self, item, variable, context=None, sort_options=None, **kwargs):
        def get_option(name):
            return self.get_option(name, context, sort_options)

        and_ = get_option('and')
        delimiter = get_option('delimiter')
        delimiter_precedes_et_al = get_option('delimiter-precedes-et-al')
        delimiter_precedes_last = get_option('delimiter-precedes-last')

        et_al_min = get_option('et-al-min')
        et_al_use_first = get_option('et-al-use-first')
        et_al_subseq_min = get_option('et-al-subsequent-min')
        et_al_subseq_use_first = get_option('et-al-subsequent-use-first')
        et_al_use_last = get_option('et-al-use-last')

        initialize_with = get_option('initialize-with')
        name_as_sort_order = get_option('name-as-sort-order')
        sort_separator = get_option('sort-separator')

        form = get_option('form')
        demote_ndp = get_option('demote-non-dropping-particle')

        def format_name_parts(given, family):
            for part in self.findall('cs:name-part', self.nsmap):
                given, family = part.format_part(given, family)
            return given, family

        names = item.reference.get(variable, [])

        if and_ == 'text':
            and_term = self.get_term('and').single
        elif and_ == 'symbol':
            and_term = self.preformat('&')

        et_al = self.et_al()
        output = []
        if form == 'count':
            count = min(len(names), et_al_use_first)
            output.append(count)
            return sum(output)
        else:
            et_al_truncate = (len(names) > 1 and et_al_min and
                              len(names) >= et_al_min)
            et_al_last = et_al_use_last and et_al_use_first <= et_al_min - 2
            if et_al_truncate:
                if et_al_last:
                    names = names[:et_al_use_first] + [names[-1]]
                else:
                    names = names[:et_al_use_first]
            for i, name in enumerate(names):
                given, family, dp, ndp, suffix = name.parts()

                if given is not None and initialize_with is not None:
                    given = self.initialize(given, initialize_with, context)

                if form == 'long':
                    if (name_as_sort_order == 'all' or
                            (name_as_sort_order == 'first' and i == 0)):
                        if demote_ndp in ('never', 'sort-only'):
                            family = ' '.join([n for n in (ndp, family) if n])
                            given = ' '.join([n for n in (given, dp) if n])
                        else:
                            given = ' '.join([n for n in (given, dp, ndp) if n])

                        given, family = format_name_parts(given, family)
                        order = family, given, suffix
                        text = sort_separator.join([n for n in order if n])
                    else:
                        family = ' '.join([n for n in (dp, ndp, family) if n])
                        given, family = format_name_parts(given, family)
                        order = given, family, suffix
                        text = ' '.join([n for n in order if n])
                elif form == 'short':
                    family = ' '.join([n for n in (ndp, family) if n])
                    given, family = format_name_parts(given, family)
                    text = family

                output.append(text)

            if et_al_truncate and et_al:
                if et_al_last:
                    ellipsis = self.unicode_character('horizontal ellipsis')
                    output[-1] = ellipsis + ' ' + output[-1]
                    text = self.join(output, delimiter)
                elif (delimiter_precedes_et_al == 'always' or
                      (delimiter_precedes_et_al == 'contextual' and
                     len(output) >= 2)):
                    output.append(et_al)
                    text = self.join(output, delimiter)
                else:
                    text = self.join(output, delimiter) + ' ' + et_al
            elif and_ is not None and len(output) > 1:
                text = self.join(output[:-1], ', ')
                if (delimiter_precedes_last == 'always' or
                    (delimiter_precedes_last == 'contextual' and
                     len(output) > 2)):
                        text = self.join([text, ''])
                else:
                    text += ' '
                text += '{} '.format(and_term) + output[-1]
            else:
                text = self.join(output, delimiter)

            return text

    def initialize(self, given, mark, context):
        if self.get_option('initialize-with-hyphen', context):
            hyphen_parts = given.split('-')
        else:
            hyphen_parts = [given.replace('-', ' ')]

        result_parts = []
        for hyphen_part in hyphen_parts:
            parts = hyphen_part.replace('.', ' ').split()
            hyphen_result = ''
            group = []
            for part in parts:
                if part[0].isupper():
                    group.append(part[0])
                else:
                    # don't initialize particles (which aren't capitalized)
                    hyphen_result += mark.join(group) + mark + ' ' + part + ' '
                    group = []
            hyphen_result += mark.join(group) + mark
            # remove double spaces
            hyphen_result = ' '.join(hyphen_result.split())
            result_parts.append(hyphen_result)
        return '-'.join(result_parts)

    def markup(self, text):
        if text:
            return self.wrap(self.format(text))
        else:
            return None


class Name_Part(CitationStylesElement, Formatted, Affixed, TextCased):
    def format_part(self, given, family):
        if self.get('name') == 'given':
            given = self.wrap(self.format(self.case(given)))
        elif self.get('name') == 'family':
            family = self.wrap(self.format(self.case(family)))
        return given, family


class Et_Al(CitationStylesElement, Formatted, Affixed):
    def process(self):
        variable = self.get('term', 'et-al')
        term = self.get_term('variable')
        return term

    def markup(self, text):
        if text:
            return self.wrap(self.format(text))
        else:
            return None


class Substitute(CitationStylesElement, Parent):
    def render(self, item, context=None, **kwargs):
        for child in self.getchildren():
            try:
                if isinstance(child, Names) and child.name is None:
                    names = self.xpath_search('./parent::cs:names[1]')[0]
                    text = child.render(item, names_context=names,
                                        context=context, **kwargs)
                else:
                    text = child.render(item, context=context, **kwargs)
            except VariableError:
                continue
            if text:
                self.add_to_repressed_list(child, context)
                break
        try:
            return text
        except NameError:
            return None

    def add_to_repressed_list(self, child, context):
        layout = context.get_layout()
        tag_list = layout.repressed.get(child.tag, [])
        tag_list.append(child.get('variable'))
        layout.repressed[child.tag] = tag_list


class Label(CitationStylesElement, Formatted, Affixed, StrippedPeriods,
            TextCased):
    def calls_variable(self):
        return self.get('variable') == 'locator'

    def process(self, item, variable=None, plural=None, context=None, **kwargs):
        if variable is None:
            variable = self.get('variable')
        form = self.get('form', 'long')
        plural_option = self.get('plural', 'contextual')
        if plural is None:
            plural = self._is_plural(item)

        if variable == 'locator' and item.has_locator:
            variable = item.locator.label

        if form == 'long':
            term = self.get_term(variable)
        else:
            term = self.get_term(variable, form)

        if (plural_option == 'contextual' and plural or
            plural_option == 'always'):
            text = term.multiple
        else:
            text = term.single

        return text

    def markup(self, text):
        if text:
            return self.wrap(self.format(self.case(self.strip_periods(text))))
        else:
            return None

    RE_MULTIPLE_NUMBERS = re.compile(r'\d+[^\d]+\d+')

    def _is_plural(self, item):
        variable = self.get('variable')
        if variable == 'locator':
            value = item.locator.identifier
        else:
            try:
                value = item.reference[variable.replace('-', '_')]
            except VariableError:
                return False

        if variable.startswith('number-of') and int(item[variable]) > 1:
            return True
        else:
            return self.RE_MULTIPLE_NUMBERS.search(str(value)) is not None


class Group(CitationStylesElement, Parent, Formatted, Affixed, Delimited):
    def calls_variable(self):
        return True

    def process(self, item, context=None, **kwargs):
        output = []
        variable_called = False
        variable_rendered = False
        for child in self.iterchildren():
            variable_called = variable_called or child.calls_variable()
            try:
                child_text = child.render(item, context=context, **kwargs)
                if child_text is not None:
                    output.append(child_text)
                    variable_rendered = (variable_rendered or
                                         child.calls_variable())
            except VariableError:
                pass
        output = [item for item in output if item is not None]
        success = not variable_called or (variable_called and variable_rendered)
        if output and success:
            return self.join(output)
        else:
            raise VariableError

    def markup(self, text):
        if text:
            return self.wrap(self.format(text))
        else:
            return None


class ConditionFailed(Exception):
    pass


class Choose(CitationStylesElement, Parent):
    def render(self, item, context=None, **kwargs):
        for child in self.getchildren():
            try:
                return child.render(item, context=context, **kwargs)
            except ConditionFailed:
                continue

        return None


class If(CitationStylesElement, Parent):
    def render(self, item, context=None, **kwargs):
        # TODO self.get('disambiguate')
        results = []
        if 'type' in self.attrib:
            results += self._type(item)
        if 'variable' in self.attrib:
            results += self._variable(item)
        if 'is-numeric' in self.attrib:
            results += self._is_numeric(item)
        if 'is-uncertain-date' in self.attrib:
            results += self._is_uncertain_date(item)
        if 'locator' in self.attrib:
            results += self._locator(item)
        if 'position' in self.attrib:
            results += self._position(item, context)

        # TODO: 'match' also applies to individual tests above!
        if self.get('match') == 'any':
            result = any(results)
        elif self.get('match') == 'none':
            result = not any(results)
        else:
            result = all(results)

        if not result:
            raise ConditionFailed

        return self.render_children(item, context=context, **kwargs)

    def _type(self, item):
        return [typ.lower() == item.reference.type
                for typ in self.get('type').split()]

    def _variable(self, item):
        variables = [var.replace('-', '_')
                     for var in self.get('variable').split()]
        output = []
        for variable in variables:
            if variable == 'locator':
                output.append('locator' in item)
            else:
                output.append(variable in item.reference)
        return output

    re_numeric = re.compile(r'^([A-Z]*\d+[A-Z]*)$', re.I)

    def _is_numeric(self, item):
        numeric_match = self.re_numeric.match
        return [var.replace('-', '_') in item.reference and
                numeric_match(str(item.reference[var.replace('-', '_')]))
                for var in self.get('is-numeric').split()]

    def _is_uncertain_date(self, item):
        result = []
        for date in self.get('is-uncertain-date').split():
            date_variable = date.replace('-', '_')
            try:
                circa = item.reference[date_variable].get('circa', False)
            except VariableError:
                circa = False
            result.append(circa)
        return result

    def _locator(self, item):
        return [var.replace('-', ' ') == item.locator.label
                for var in self.get('locator').split()]

    def _position(self, item, context):
        if context is None:
            context = self
        if context.xpath_search('./ancestor::*[self::cs:bibliography]'):
            return [False]
        # citation node
        cites = context.get_layout().getparent().cites
        last_cite = cites[-1] if cites else None
        already_cited = item.key in (cite.key for cite in cites)
        possibly_ibid = (already_cited
                         and item.key == last_cite.key
                         and (item.citation is last_cite.citation
                              or len(last_cite.citation.cites) == 1))
        results = []
        for position in self.get('position').split():
            result = False
            if position == 'first':
                result = not already_cited
            elif position == 'subsequent':
                result = already_cited
            elif possibly_ibid and position == 'ibid':
                result = item.has_locator or not last_cite.has_locator
            elif possibly_ibid and position == 'ibid-with-locator':
                result = (item.has_locator and not last_cite.has_locator
                          or (item.has_locator and last_cite.has_locator
                              and item.locator != last_cite.locator))
            elif already_cited and position == 'near-note':
                max_distance = self.get_root().get_option('near-note-distance')
                citations = 1
                last_citation = None
                for cite in reversed(cites):
                    if item.key == cite.key and citations <= max_distance:
                        result = True
                        break
                    elif cite.citation is not last_citation:
                        citations += 1
                        last_citation = cite.citation
            results.append(result)
        return results


class Else_If(If, CitationStylesElement):
    pass


class Else(CitationStylesElement, Parent):
    def render(self, item, context=None, **kwargs):
        return self.render_children(item, context=context, **kwargs)


# utility functions

def to_ordinal(number, context):
    number = str(number)
    last_digit = int(number[-1])
    if last_digit in (1, 2, 3) and not (len(number) > 1 and number[-2] == '1'):
        ordinal_term = 'ordinal-{:02}'.format(last_digit)
    else:
        ordinal_term = 'ordinal-04'
    return number + context.get_term(ordinal_term).single


def romanize(n):
    # by Kay Schluehr - from http://billmill.org/python_roman.html
    numerals = (('M', 1000), ('CM', 900), ('D', 500), ('CD', 400),
                ('C', 100),('XC', 90),('L', 50),('XL', 40), ('X', 10),
                ('IX', 9), ('V', 5), ('IV', 4), ('I', 1))
    roman = []
    for ltr, num in numerals:
        (k, n) = divmod(n, num)
        roman.append(ltr * k)
    return ''.join(roman)




import glob
import io
import os
import sys
import traceback

from codecs import utf_8_encode
from functools import reduce
from optparse import OptionParser
from subprocess import Popen, PIPE

from citeproc import CitationStylesStyle, CitationStylesBibliography
from citeproc.source import Citation, CitationItem, Locator
from citeproc.source.json import CiteProcJSON


CITEPROC_TEST_REPOSITORY = 'https://github.com/citation-style-language/test-suite.git'
CITEPROC_TEST_COMMIT = 'c3db429ab7c6b9b9ccaaa6d3c6bb9e503f0d7b11'

CITEPROC_TEST_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                     'test-suite'))
TEST_PARSER_PATH = os.path.join(CITEPROC_TEST_PATH, 'processor.py')
TESTS_PATH = os.path.join(CITEPROC_TEST_PATH, 'processor-tests', 'humans')

FAILING_TESTS_FILE = 'failing_tests.txt'


# The results of the following tests are ignored, since they don't test CSL
# features, but citeproc-js specific features

IGNORED_RESULS = {
    'date_Accessed': 'raw date',
    'date_LopsidedDataYearSuffixCollapse': 'raw date',
    'date_SeasonRange1': 'raw date',
    'date_SeasonRange2': 'raw date',
    'date_SeasonRange3': 'raw date',
    'date_String': 'raw date',
}


class ProcessorTest(object):
    """Parses atest fixture and provides a method for processing the tests
    defined in it."""
    bib_prefix = '<div class="csl-bib-body">'
    bib_suffix = '</div>'
    item_prefix = '  <div class="csl-entry">'
    item_suffix = '</div>'

    def __init__(self, filename):
        self.csl_test = CslTest({}, None, (filename, ),
                                os.path.basename(filename))
        self.csl_test.parse()
        csl_io = io.BytesIO(utf_8_encode(self.data['csl'])[0])
        self.style = CitationStylesStyle(csl_io)
        self._fix_input(self.data['input'])
        self.references = [item['id'] for item in self.data['input']]
        self.references_dict = CiteProcJSON(self.data['input'])
        self.bibliography = CitationStylesBibliography(self.style,
                                                       self.references_dict)
        self.expected = self.data['result'].splitlines()

    @property
    def data(self):
        return self.csl_test.data

    @staticmethod
    def _fix_input(input_data):
        for i, ref in enumerate(input_data):
            if 'id' not in ref:
                ref['id'] = i
            if 'type' not in ref:
                ref['type'] = 'book'

    def execute(self):
        if self.data['citation_items']:
            citations = [self.parse_citation(item)
                         for item in self.data['citation_items']]
        elif self.data['citations']:
            citations = []
            for cit in self.data['citations']:
                cit = cit[0]
                citation_items = [self.parse_citation_item(cititem)
                                  for cititem in cit['citationItems']]
                citation = Citation(citation_items)
                citation.key = cit['citationID']
                citation.note_index = cit['properties']['noteIndex']
                citations.append(citation)
        elif self.data['bibentries']:
            citation_items = [self.parse_citation_item({'id': entry})
                              for entry in self.data['bibentries'][-1]]
            citations = [Citation(citation_items)]
        else:
            citation_items = [self.parse_citation_item({'id': ref})
                              for ref in self.references]
            citations = [Citation(citation_items)]

        for citation in citations:
            self.bibliography.register(citation)

        if self.style.has_bibliography():
            self.bibliography.sort()

        results = []
        do_nothing = lambda x: None     # callback passed to cite()
        if self.data['mode'] == 'citation':
            if self.data['citations']:
                for i, citation in enumerate(citations):
                    if i == len(citations) - 1:
                        dots_or_other = '>>'
                    else:
                        dots_or_other = '..'
                    results.append('{}[{}] '.format(dots_or_other, i) +
                                   self.bibliography.cite(citation, do_nothing))
            else:
                for citation in citations:
                    results.append(self.bibliography.cite(citation, do_nothing))
        elif self.data['mode'] in ('bibliography', 'bibliography-nosort'):
            results.append(self.bib_prefix)
            for entry in self.bibliography.bibliography():
                text = self.item_prefix + str(entry) + self.item_suffix
                results.append(text)
            results.append(self.bib_suffix)
        return results

    def parse_citation(self, citation_data):
        citation_items = []
        for item in citation_data:
            citation_item = self.parse_citation_item(item)
            citation_items.append(citation_item)
        return Citation(citation_items)

    def parse_citation_item(self, citation_item_data):
        options = {}
        for key, value in citation_item_data.items():
            python_key = key.replace('-', '_')
            if python_key == 'id':
                reference_key = str(value)
                continue
            elif python_key == 'locator':
                try:
                    options['locator'] = Locator(citation_item_data['label'],
                                                 value)
                except KeyError:
                    # some tests don't specify the label
                    options['locator'] = Locator('page', value)
            elif python_key == 'label':
                pass
            else:
                options[python_key] = value
        return CitationItem(reference_key, **options)


class FailedTests(object):
    """Read the known failed tests from a file and update the file with the
    results from a new test run."""
    def __init__(self, filename):
        self.filename = filename
        with open(filename, 'r') as file:
            self.lines = file.readlines()
        self.now_failing = {}
        self.updated_lines = []

    def mark_failure(self, test_name, reason=None):
        self.now_failing[test_name] = reason

    def compare(self):
        assert not self.updated_lines
        was_failing = set()
        new_failing_tests = []
        new_fixed_tests = []
        for line in self.lines:
            test_name, comment = (line.split('#') if '#' in line
                                  else (line, None))
            test_name = test_name.strip()
            if test_name and test_name not in self.now_failing:
                new_fixed_tests.append(test_name)
                continue
            was_failing.add(test_name)
            self.updated_lines.append(line)
        for test_name in sorted(self.now_failing.keys()):
            if test_name not in was_failing:
                reason = self.now_failing[test_name]
                line = ('{:<66} # {}'.format(test_name, reason) if reason
                        else test_name)
                self.updated_lines.append(line + '\n')
                new_failing_tests.append(test_name)
        return new_failing_tests, new_fixed_tests

    def update_file(self):
        assert self.updated_lines
        with open(self.filename, 'w') as file:
            file.writelines(self.updated_lines)


def execute_git_command(args, echo=False):
    """Execute a git command, exiting this script when it returns a
    non-zero error code."""
    command = ['git'] + args
    if echo:
        print(' '.join(command))
    process = Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    out, err = process.communicate()
    if process.returncode:
        print('Calling git failed. Make sure git is installed. Aborting.')
        sys.exit(1)
    return out, err


def clone_test_suite():
    """Clone the citeproc-test repository if it is not present. Otherwise, check
    whether the correct commit is checked out."""
    if not os.path.exists(CITEPROC_TEST_PATH):
        git_clone = ['clone', CITEPROC_TEST_REPOSITORY, CITEPROC_TEST_PATH]
        print('Cloning the test-suite repository...')
        execute_git_command(git_clone, echo=True)
        git_checkout = ['-C', CITEPROC_TEST_PATH, 'checkout',
                        CITEPROC_TEST_COMMIT]
        print('Checking out the specified commit...')
        execute_git_command(git_checkout)
    else:
        git_ref = ['-C', CITEPROC_TEST_PATH, 'rev-parse', 'HEAD']
        commit_ref, _ = execute_git_command(git_ref)
        if commit_ref.strip() != CITEPROC_TEST_COMMIT.encode('ascii'):
            print('The checked-out commit of test-suite does not match \n'
                  'the one recorded in this test script. Aborting.')
            sys.exit(1)
        git_fetch = ['-C', CITEPROC_TEST_PATH, 'fetch']
        execute_git_command(git_fetch)
    git_ref = ['-C', CITEPROC_TEST_PATH, 'rev-parse', 'origin/master']
    commit_ref, _ = execute_git_command(git_ref)
    has_updates = commit_ref.strip() != CITEPROC_TEST_COMMIT.encode('ascii')
    return has_updates


RED = '\033[91m'
GREEN = '\033[92m'
BOLD = '\033[1m'
END = '\033[0m'


if __name__ == '__main__':
    usage = ('usage: %prog [options] glob_pattern\n\n'
             'glob_pattern limits the tests that are executed, for example:\n'
             '  %prog *Sort*\n'
             'runs only test fixtures that have "Sort" in the filename')
    parser = OptionParser(usage)
    parser.add_option('-m', '--max', dest='max', default=-1,
                      help='run up to MAX tests', metavar='MAX')
    parser.add_option('-v', '--verbose', dest='verbose', action='store_true',
                      default=False, help='print results for each test')
    parser.add_option('-r', '--raise', dest='catch_exceptions', default=True,
                      action='store_false',
                      help='do not catch exceptions (aborts program)')
    parser.add_option('-s', '--summary', dest='summary', action='store_true',
                      default=False, help='print a summary of the test results')
    parser.add_option('-n', '--no-update', dest='update', action='store_false',
                      default=True, help='do not update {}'
                                         .format(FAILING_TESTS_FILE))
    parser.add_option('-f', '--file', dest='file', default=None,
                      help='write output to FILE', metavar='FILE')
    (options, args) = parser.parse_args()

    max_tests = int(options.max)
    try:
        destination = open(options.file, 'wt', encoding='utf-8')
        class UnicodeWriter(object):
            def write(self, s):
                destination.write(str(s))
        sys.stderr = UnicodeWriter()
    except TypeError:
        destination = sys.stdout

    def out(*args):
        if not args:
            destination.write('\n')
        else:
            print(*args, file=destination)

    def print_result(name, passed, total):
        out(' {:<13} {:>3} / {:>3} ({:>4.0%})'.format(name, passed, total,
                                                      passed / total))

    try:
        glob_pattern = args[0]
        run_all_tests = False
    except IndexError:
        glob_pattern = '*'
        run_all_tests = True

    test_repo_has_updates = clone_test_suite()

    # import the text fixture parser included with citeproc-test
    try:  # Python 3.3+
        from importlib.machinery import SourceFileLoader
        loader = SourceFileLoader('processor', TEST_PARSER_PATH)
        processor = loader.load_module()
    except ImportError:  # older Python versions
        from imp import load_source
        processor = load_source('processor', TEST_PARSER_PATH)
    from processor import CslTest
    global CslTest

    total_count = {}
    passed_count = {}
    failed_tests = FailedTests(os.path.join(os.path.dirname(__file__),
                                            FAILING_TESTS_FILE))
    count = 0
    test_files = os.path.join(TESTS_PATH, '{}.txt'.format(glob_pattern))
    for filename in sorted(glob.glob(test_files)):
        test_name, _ = os.path.basename(filename).split('.txt')
        category, _ = os.path.basename(filename).split('_')
        passed_count.setdefault(category, 0)
        if count == max_tests:
            break

        reason = None
        try:
            if test_name not in IGNORED_RESULS:
                total_count[category] = total_count.get(category, 0) + 1
                count += 1
            t = ProcessorTest(filename)
            if run_all_tests and (t.data['mode'] == 'bibliography-header' or
                                  t.data['bibsection']):
                continue
            if options.verbose:
                out('>>> Testing {}'.format(os.path.basename(filename)))
                out('EXP: ' + '\n     '.join(t.expected))

            results = t.execute()
            results = reduce(lambda x, y: x+y,
                             [item.split('\n') for item in results])
            results = [item.replace('&amp;', '&#38;')
                       for item in results]
            if options.verbose:
                out('RES: ' + '\n     '.join(results))
            if results == t.expected:
                if test_name not in IGNORED_RESULS:
                    passed_count[category] += 1
                if options.verbose:
                    out('<<< SUCCESS\n')
                continue
            else:
                if options.verbose:
                    out('<<< FAILED\n')
            del t
        except Exception as e:
            if not options.catch_exceptions:
                raise
            if options.verbose:
                out('Exception in', os.path.basename(filename))
                traceback.print_exc()
                out('<<< FAILED\n')
            reason = 'uncaught exception'
        if test_name not in IGNORED_RESULS:
            failed_tests.mark_failure(test_name, reason)

    success = False
    if sum(total_count.values()) == 0:
        print('<no tests found>: check README.md file for instructions')
    else:
        if run_all_tests:
            new_failing, new_fixed = failed_tests.compare()
            success = not (new_failing or new_fixed)
            if options.update:
                failed_tests.update_file()

            if new_failing:
                out(BOLD + 'New failing tests:' + END)
                for test_name in new_failing:
                    out(' ' + RED + test_name + END)
            if new_failing or new_fixed:
                out()
            else:
                out('All is well.')
            if new_fixed:
                out(BOLD + 'Fixed tests:' + END)
                for test_name in new_fixed:
                    out(' ' + GREEN + test_name + END)
                out()
                out('You need to fix the newly failing tests and may commit the \n'
                    'removal of the fixed tests from {}'.format(FAILING_TESTS_FILE))

            if test_repo_has_updates:
                out()
                out(BOLD + 'The test-suite repository has updates! Consider '
                           'updating the test script.' + END)
        else:
            failing = sorted(failed_tests.now_failing.keys())
            success = not failing
            out(BOLD + 'Failing tests:' + END)
            for test_name in failing:
                out(' ' + RED + test_name + END)
        if options.summary:
            out()
            out(BOLD + 'Summary:' + END)
            for category in sorted(total_count.keys()):
                print_result(category, passed_count[category],
                             total_count[category])
            out()
            print_result('total', sum(passed_count.values()),
                         sum(total_count.values()))
    try:
        destination.close()
    except AttributeError:
        pass
    sys.exit(0 if success else 1)



# coding: utf-8

import unicodedata

from . import (parse_argument, eat_whitespace, parse_macro_name,
               OPEN_SCOPE, CLOSE_SCOPE, START_MACRO)


__all__ = ['MACROS', 'NewCommand', 'Macro']


class MacroBase(object):
    def parse_arguments(self, tokens):
        raise NotImplementedError

    def parse_arguments_and_expand(self, tokens):
        raise NotImplementedError



class NewCommand(MacroBase):
    r""" \newcommand{cmd}[args]{def} """

    def __init__(self, macros):
        self.macros = macros

    @staticmethod
    def _parse_macro_name(tokens):
        eat_whitespace(tokens)
        token = next(tokens)
        in_group = token.type == OPEN_SCOPE
        if in_group:
            eat_whitespace(tokens)
            token = next(tokens)
        assert token.type == START_MACRO
        name = parse_macro_name(tokens)
        if in_group:
            eat_whitespace(tokens)
            assert next(tokens).type == CLOSE_SCOPE
        return name

    @staticmethod
    def _parse_optional_arguments(tokens, macros):
        eat_whitespace(tokens)
        if tokens.peek().value == '[':
            next(tokens)
            num_args = ''
            for token in tokens:
                if token.value == ']':
                    break
                num_args += token.value
        else:
            num_args = 0
        return int(num_args)

    def parse_arguments_and_expand(self, tokens, macros):
        name = self._parse_macro_name(tokens)
        num_args = self._parse_optional_arguments(tokens, macros)
        definition = parse_argument(tokens, macros)
        for i in range(10):
            definition = definition.replace('#{}'.format(i + 1),
                                            '{' + str(i) + '}')
        self.macros[name] = Macro(num_args, definition)
        return ''


class Macro(object):
    def __init__(self, num_args, format_string):
        self.num_args = num_args
        self.format_string = format_string

    def parse_arguments_and_expand(self, tokens, macros):
        args = [parse_argument(tokens, macros) for _ in range(self.num_args)]
        return self.expand(args)

    def expand(self, arguments):
        assert len(arguments) == self.num_args
        return self.format_string.format(*arguments)


class Symbol(Macro):
    def __init__(self, symbol):
        super(Symbol, self).__init__(0, symbol)


class SymbolByName(Macro):
    def __init__(self, unicode_symbol_name):
        unicode_symbol = unicodedata.lookup(unicode_symbol_name)
        super(SymbolByName, self).__init__(0, unicode_symbol)


class Combining(Macro):
    def __init__(self, unicode_accent_name):
        unicode_accent = unicodedata.lookup('COMBINING ' + unicode_accent_name)
        super(Combining, self).__init__(1, '{0}' + unicode_accent)

    DOTTED_CHARS = {'ı': 'i',
                    'ȷ': 'j'}

    def expand(self, arguments):
        assert len(arguments) == self.num_args
        accented, rest = arguments[0][0], arguments[0][1:]
        accented = self.DOTTED_CHARS.get(accented, accented)
        expanded = super(Combining, self).expand([accented])
        return unicodedata.normalize('NFC', expanded) + rest


MACROS = {
    # accents
    '`': Combining('GRAVE ACCENT'),
    "'": Combining('ACUTE ACCENT'),
    '^': Combining('CIRCUMFLEX ACCENT'),
    '"': Combining('DIAERESIS'),
    'H': Combining('DOUBLE ACUTE ACCENT'),
    '~': Combining('TILDE'),
    'c': Combining('CEDILLA'),
    'k': Combining('OGONEK'),
    '=': Combining('MACRON'),
    'b': Combining('MACRON BELOW'),
    '.': Combining('DOT ABOVE'),
    'd': Combining('DOT BELOW'),
    'r': Combining('RING ABOVE'),
    'u': Combining('BREVE'),
    'v': Combining('CARON'),
    # '|': Combining('VERTICAL LINE ABOVE'),
    # 'h': Combining('HOOK ABOVE'),
    'G': Combining('DOUBLE GRAVE ACCENT'),
    'U': Combining('DOUBLE VERTICAL LINE ABOVE'),
    't': Combining('DOUBLE INVERTED BREVE'),
    'textcircled': Combining('ENCLOSING CIRCLE'),

    # symbols
    'o': SymbolByName('LATIN SMALL LETTER O WITH STROKE'),
    'O': SymbolByName('LATIN CAPITAL LETTER O WITH STROKE'),
    'i': SymbolByName('LATIN SMALL LETTER DOTLESS I'),
    'l': SymbolByName('LATIN SMALL LETTER L WITH STROKE'),
    'L': SymbolByName('LATIN CAPITAL LETTER L WITH STROKE'),
    'oe': SymbolByName('LATIN SMALL LIGATURE OE'),
    'OE': SymbolByName('LATIN CAPITAL LIGATURE OE'),
    'ae': SymbolByName('LATIN SMALL LETTER AE'),
    'AE': SymbolByName('LATIN CAPITAL LETTER AE'),
    'aa': SymbolByName('LATIN SMALL LETTER A WITH RING ABOVE'),
    'AA': SymbolByName('LATIN CAPITAL LETTER A WITH RING ABOVE'),
    'ss': SymbolByName('LATIN SMALL LETTER SHARP S'),
    'dh': SymbolByName('LATIN SMALL LETTER ETH'),
    'DH': SymbolByName('LATIN CAPITAL LETTER ETH'),
    'dj': SymbolByName('LATIN SMALL LETTER D WITH STROKE'),
    'DJ': SymbolByName('LATIN CAPITAL LETTER D WITH STROKE'),
    'ng': SymbolByName('LATIN SMALL LETTER ENG'),
    'NG': SymbolByName('LATIN CAPITAL LETTER ENG'),
    'th': SymbolByName('LATIN SMALL LETTER THORN'),
    'TH': SymbolByName('LATIN CAPITAL LETTER THORN'),
    'dag': SymbolByName('DAGGER'),
    'ddag': SymbolByName('DOUBLE DAGGER'),
    'dots': SymbolByName('HORIZONTAL ELLIPSIS'),
    'P': SymbolByName('PILCROW SIGN'),
    'S': SymbolByName('SECTION SIGN'),
    'copyright': SymbolByName('COPYRIGHT SIGN'),
    'pounds': SymbolByName('POUND SIGN'),
    'guillemotleft': SymbolByName('LEFT-POINTING DOUBLE ANGLE QUOTATION MARK'),
    'guillemotright': SymbolByName('RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK'),
    'guilsinglleft': SymbolByName('SINGLE LEFT-POINTING ANGLE QUOTATION MARK'),
    'guilsinglright': SymbolByName('SINGLE RIGHT-POINTING ANGLE QUOTATION MARK'),
    'quotedblbase': SymbolByName('DOUBLE LOW-9 QUOTATION MARK'),
    'quotesinglbase': SymbolByName('SINGLE LOW-9 QUOTATION MARK'),

    'textasciicircum': Symbol('^'),
    'textasciitilde': Symbol('~'),
    'textasteriskcentered': Symbol('*'),
    'textbackslash': Symbol('\\'),
    'textbar': Symbol('|'),
    'textbraceleft': Symbol('{'),
    'textbraceright': Symbol('}'),
    'textbullet': SymbolByName('BULLET'),
    'textcopyright': SymbolByName('COPYRIGHT SIGN'),
    'textdagger': SymbolByName('DAGGER'),
    'textdaggerdbl': SymbolByName('DOUBLE DAGGER'),
    'textdollar': Symbol('$'),
    'textellipsis': SymbolByName('HORIZONTAL ELLIPSIS'),
    'textemdash': SymbolByName('EM DASH'),
    'textendash': SymbolByName('EN DASH'),
    'textexclamdown': SymbolByName('INVERTED EXCLAMATION MARK'),
    'textgreater': Symbol('>'),
    'textless': Symbol('<'),
    'textordfeminine': SymbolByName('FEMININE ORDINAL INDICATOR'),
    'textordmasculine': SymbolByName('MASCULINE ORDINAL INDICATOR'),
    'textparagraph': SymbolByName('PILCROW SIGN'),
    'textperiodcentered': SymbolByName('MIDDLE DOT'),
    'textquestiondown': SymbolByName('INVERTED QUESTION MARK'),
    'textquotesingle': Symbol("'"),  # from the textcomp package
    'textquotedbl': Symbol('"'),
    'textquotedblleft': SymbolByName('LEFT DOUBLE QUOTATION MARK'),
    'textquotedblright': SymbolByName('RIGHT DOUBLE QUOTATION MARK'),
    'textquoteleft': SymbolByName('LEFT SINGLE QUOTATION MARK'),
    'textquoteright': SymbolByName('RIGHT SINGLE QUOTATION MARK'),
    'textregistered': SymbolByName('REGISTERED SIGN'),
    'textsection': SymbolByName('SECTION SIGN'),
    'textsterling': SymbolByName('POUND SIGN'),
    'texttrademark': SymbolByName('TRADE MARK SIGN'),
    'textunderscore': Symbol('_'),
    'textvisiblespace': SymbolByName('OPEN BOX'),

    'TeX': Macro(0, 'TeX'),

    # escaped characters
    ' ': Symbol(' '),
    '&': Symbol('&'),
    '$': Symbol('$'),
    '{': Symbol('{'),
    '}': Symbol('}'),
    '%': Symbol('%'),
    '#': Symbol('#'),
    '_': Symbol('_'),
}




import unicodedata

from collections import namedtuple
from warnings import warn


__all__ = ['parse_latex', 'substitute_ligatures']


def parse_latex(string, macros={}):
    tokens = Tokenizer(string)
    output = ''
    for result in dispatch(tokens, macros):
        output += result
    return substitute_ligatures(output)


Token = namedtuple('Token', ['type', 'value'])


TOGGLE_MATH = 'TOGGLE-MATH'
WHITESPACE = 'WHITESPACE'
CLOSE_SCOPE = 'CLOSE-SCOPE'
OPEN_SCOPE = 'OPEN-SCOPE'
START_MACRO = 'START-MACRO'
CHARACTER = 'CHARACTER'


class Tokenizer(object):
    def __init__(self, string):
        self.string = string
        self._tokens = self.tokenize(string)
        self._next_token = None

    @staticmethod
    def tokenize(input_string):
        for char in input_string:
            if char == '\\':
                yield Token(START_MACRO, char)
            elif char == '{':
                yield Token(OPEN_SCOPE, char)
            elif char == '}':
                yield Token(CLOSE_SCOPE, char)
            elif char in ' \t\n':
                yield Token(WHITESPACE, char)
            elif char == '$':
                yield Token(TOGGLE_MATH, char)
            else:
                yield Token(CHARACTER, char)

    def __iter__(self):
        return self

    def __next__(self):
        if self._next_token:
            token = self._next_token
            self._next_token = None
        else:
            token = next(self._tokens)
        return token

    next = __next__

    def peek(self):
        if self._next_token is None:
            self._next_token = next(self._tokens)
        return self._next_token


def eat_whitespace(tokens):
    while tokens.peek().type == WHITESPACE:
        next(tokens)


class ScopeClosing(Exception):
    pass


def dispatch(tokens, macros, level=0):
    while True:
        try:
            next_token = tokens.peek()
        except StopIteration:
            if level > 0:
                warn("Unbalanced parenthesis in '{}'".format(tokens.string))
            break
        if next_token.type == OPEN_SCOPE:
            yield handle_scope(tokens, macros, level)
        elif next_token.type == CLOSE_SCOPE:
            raise ScopeClosing
        elif next_token.type == START_MACRO:
            yield handle_macro(tokens, macros)
        elif next_token.type == TOGGLE_MATH:
            yield handle_math(tokens)
        else:
            assert next_token.type in (CHARACTER, WHITESPACE)
            yield next(tokens).value


def handle_scope(tokens, macros, level):
    assert next(tokens).type == OPEN_SCOPE
    output = ''
    try:
        for result in dispatch(tokens, macros, level + 1):
            output += result
    except ScopeClosing:
        assert next(tokens).type == CLOSE_SCOPE
    return output


def parse_argument(tokens, macros, level=0):
    eat_whitespace(tokens)
    return next(dispatch(tokens, macros, level))


def handle_macro(tokens, macros):
    assert next(tokens).type == START_MACRO
    name = parse_macro_name(tokens)
    try:
        macro = MACROS[name]
    except KeyError:
        macro = macros[name]
    return macro.parse_arguments_and_expand(tokens, macros)


def parse_macro_name(tokens):
    token = next(tokens)
    assert token.type in (CHARACTER, TOGGLE_MATH, WHITESPACE)
    name = token.value
    if name.isalpha():
        while tokens.peek().type == CHARACTER and tokens.peek().value.isalpha():
            name += next(tokens).value
        eat_whitespace(tokens)
    return name


def handle_math(tokens):
    assert next(tokens).type == TOGGLE_MATH
    output = ''
    for token in tokens:
        if token.type == START_MACRO:
            output += token.value
            token = next(tokens)
        elif token.type == TOGGLE_MATH:
            break
        output += token.value
    return '$' + output + '$'


def substitute_ligatures(string):
    for chars, ligature in SUBSTITUTIONS.items():
        string = string.replace(chars, unicodedata.lookup(ligature))
    return string


SUBSTITUTIONS = {"~": 'NO-BREAK SPACE',

                 # ligatures defined in Computer Modern (symbol shortcuts)
                 "--": 'EN DASH',
                 "---": 'EM DASH',
                 "''": 'RIGHT DOUBLE QUOTATION MARK',
                 "``": 'LEFT DOUBLE QUOTATION MARK',
                 "!`": 'INVERTED EXCLAMATION MARK',
                 "?`": 'INVERTED QUESTION MARK',
                 ",,": 'DOUBLE LOW-9 QUOTATION MARK',
                 "<<": 'LEFT-POINTING DOUBLE ANGLE QUOTATION MARK',
                 ">>": 'RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK',
}


from .macro import MACROS
